/*
 * PredErrorCoder.c
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Contains:
 *    Higher-level functions for prediction error coding
 *    in INTER frame. 
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "mainFlags.h"
#include "mcp.h"
#include "encMux.h"
#include "decDemux.h"
#include "tools.h"
#include "deblockFilt.h"
#include "PredErrorCoder.h"
#include "PredErrorDecoder.h"
#include "encMCC.h"
#include "decMCC.h"

#define MIN_BITS_8x8Y              4
#define MIN_BITS_8x8C              4
#define MIN_BITS_4x4C              3
#define MIN_IMPROVEMENT          100
extern int
  tableInter[65],
  tableIntra[65];

extern u_char 
  zig4x4[16],
  zig8x8[64];

extern int *region_intra_dc;
extern triplet *region_triplets;
extern doublet *region_doublets;
extern triplet *vq_triplets;
extern int *rdc_value;

extern int     
  triplet_counter,
  intra_dc_counter,
  doublet_counter,
  vq_counter,
  rdc_counter;

extern u_char 
  annexTlookup[32];

extern int
  triplet_counter_Y,
  intra_dc_counter_Y,
  doublet_counter_Y,
  vq_counter_Y,
  rdc_counter_Y;


/* Encoder vlc tables for region mode */

vlc_item

  /* Table for 16x16 region when motion prediction is not possible */
  vlcEncRegionModeNoPred16[6] = {
    { 3,       3},
    { 4,       1},
    { 1,       1},
    { 3,       1},
    { 4,       0},
    { 3,       2},
  },

  /* Table for 16x16 region when motion prediction is possible */
  vlcEncRegionModePred16[10] = {
    { 4,      10},
    { 4,       9},
    { 2,       1},
    { 4,      11},
    { 4,      12},
    { 4,      13},
    { 3,       7},
    { 5,      17},
    { 5,      16},
    { 2,       0},
  },

  /* Table for 8x8 region when motion prediction is not possible */
  vlcEncRegionModeNoPred8[5] = {
    { 3,       3},
    { 3,       1},
    { 1,       1},
    { 3,       2},
    { 3,       0},
  },

  /* Table for 8x8 region when motion prediction is possible */
  vlcEncRegionModePred8[9] = {
    { 3,       0},
    { 3,       3},
    { 3,       7},
    { 3,       4},
    { 3,       2},
    { 3,       5},
    { 3,       6},
    { 4,       3},
    { 4,       2},
  };

static long
SendMccRegion(struct frame *origfr, struct frame *predfr, int blockInd,
              struct frame *recofr, object *region, int sizex, int sizey,
              int regionSize, int regionMode,
              int predPossible, int predDirsAvail,
              int predDir, int predSelNeeded,
              int predSel, int *motCoefs, u_char *pecMode,
              int qp, int aQP, float lambda_Y,
              int numActiveRefFrames,
              int *bits_reg_mcc, int *bits_reg_mcc_uv,
              int *bits_reg_ovh, int *bits_reg_mode, int *bits_motion,
              int gen_bitstr_flag, int complexityMode);

int SendRegionMode(int regionMode, int pecMode, int predPossible, int regSize,
                   int genBitstrFlag)
{
  int
    index, pecModeIndex, bits;

  vlc_item 
    *vlcTable;

  /* 
   * Choose VLC table for region mode.
   */
  if(predPossible)
    vlcTable = (regSize==16)?vlcEncRegionModePred16:vlcEncRegionModePred8;
  else
    vlcTable = (regSize==16)?vlcEncRegionModeNoPred16:vlcEncRegionModeNoPred8;

  /* 
   * Find the index of the codeword in table.
   */
  pecModeIndex = (pecMode == PECMODE_CODED) ? 1 : 0;

  if(predPossible){
    switch(regionMode){
    case REGMODE_INTER_NOPRED:     index = 0 + pecModeIndex; break;
    case REGMODE_INTER_PRED:       index = 2 + pecModeIndex; break;
    case REGMODE_INTER_PRED_DELTA: index = 4 + pecModeIndex; break;
    case REGMODE_INTER_COPY:       index = 6 + pecModeIndex; break;
    case REGMODE_INTRA:            index = 8; break;
    case REGMODE_SPLITTED:         index = 9; break;
    }
  }
  else{
    switch(regionMode){
    case REGMODE_INTER_NOPRED:     index = 0 + pecModeIndex; break;
    case REGMODE_INTER_COPY:       index = 2 + pecModeIndex; break;
    case REGMODE_INTRA:            index = 4; break;
    case REGMODE_SPLITTED:         index = 5; break;
    }
  }

  /*
   * Send the codeword or just find the
   * amonut of bits for signaling.
   */
  if (genBitstrFlag)
    bits = put_vlc_item2(vlcTable[index], 0);
  else
    bits = vlcTable[index].len;

  return bits;
}


void store_8x8_info1(int codingMethod, int class, int *dctCoeffs, int mode)
{
  /*
   * Store 8x8 coding method and class of the 8x8 block
   * for bitstream dumping.
   */
  region_doublets[doublet_counter].transform = codingMethod;
  region_doublets[doublet_counter].class = class;
  doublet_counter++;
}


void store_8x8_info2(int *methCoeffs, int codingMethod)
{
  int 
    i, 
    codedBlocks, 
    startCoeff = 0;

  /*
   * Evaluate the number of 4x4 blocks.
   */
  codedBlocks = codingMethod == METH_4x4_15 ? 4 : 1;

  /*
   * Store 4x4 coding method and clas of the 4x4 block
   * for bitstream dumping.
   */
  for (i = 0; i < codedBlocks; i++, startCoeff += SIZE_4x4) {
    vq_triplets[vq_counter].last = methCoeffs[startCoeff + 13];
    vq_triplets[vq_counter].run = methCoeffs[startCoeff + 10];
    vq_counter++;
  }
}


void store_8x8_info3(int *methCoeffs, int codingMethod, int chroma, int mode)
{
  int 
    size, num = 0, startCoeff = 0, 
    codedCoeffs, coeff, run;

  /*
   * Evaluate the transform size from coding method.
   */
  switch(codingMethod) {

    case METH_8x8_DCT: case METH_DCT_8x8_N_0_8:
    case METH_DCT_8x8_N_4_8:
    case METH_KLT_8x8_N_1_8: case METH_KLT_8x8_N_2_8: 
    case METH_KLT_8x8_N_3_8: case METH_KLT_8x8_N_5_8:
    case METH_KLT_8x8_N_6_8: case METH_KLT_8x8_N_7_8:
      size = 64;
      break;

    case METH_4x8_DCT_1: case METH_4x8_DCT_2:
    case METH_4x8_DCT_3: case METH_4x8_DCT_4:
    case METH_8x4_DCT_1: case METH_8x4_DCT_2:
    case METH_8x4_DCT_3: case METH_8x4_DCT_4:
      size = 32;
      break;

    default:
      size = 16;
      break;
  }

  /*
   * Scan the transforms of the coding method.
   */
  while (num * size < 64) {

    /*
     * Calculate the non-zero transform coefficients.
     */
    codedCoeffs = coeff = 0; 
    while (coeff < size) {
      if (methCoeffs[startCoeff + coeff]) 
        codedCoeffs++;
      coeff++;
    }

    /*
     * Store the Run-Length coded non-zero coefficients
     * for bitstream dumping.
     */
    coeff = run = 0;
    while (codedCoeffs > 0) {
      while (methCoeffs[startCoeff + coeff] == 0) { 
        coeff++; 
        run++; 
      }
      region_triplets[triplet_counter].run = run;
      region_triplets[triplet_counter].last = codedCoeffs > 1 ? 0 : 1;
      region_triplets[triplet_counter].level = methCoeffs[startCoeff + coeff];
      triplet_counter++;
      coeff++; 
      run = 0; 
      codedCoeffs--;
    }
    num++;
    startCoeff += size;
  }
}


int SendMotionPredInfo(int predDirsAvail, int predDir,
                       int predSelNeeded, int predSel,
                       int genBitstrFlag)
{
  int 
    bits = 0;

  /*
   * Send information about the motion prediction direction
   * if needed.
   */
  if(predDirsAvail == PREDDIR_BOTH){
    bits++;
    if(genBitstrFlag){
      if(predDir == PREDDIR_UP)
        put_bits(1, 0, 0, 1);
      else
        put_bits(1, 1, 0, 1);
    }
  }

  /*
   * Send information about the reference block selection
   * if needed.
   */
  if(predSelNeeded == PREDSEL_NEEDED){
    bits++;
    if(genBitstrFlag){
      if(predSel == PREDSEL_1)
        put_bits(1, 0, 0, 1);
      else
        put_bits(1, 1, 0, 1);
    }
  }

  return bits;
}


/* qpBlockLuma used to pass qpY to vdxGetDctTable() and vdxGetKltTable() */
extern int
  qpBlockLuma;

static long
SendMccRegion(struct frame *origfr, struct frame *predfr, int blockInd,
              struct frame *recofr, object *region, int sizex, int sizey,
              int regionSize, int regionMode,
              int predPossible, int predDirsAvail,
              int predDir, int predSelNeeded,
              int predSel, int *motCoefs, u_char *pecMode,
              int qp, int aQP, float lambda_Y,
              int numActiveRefFrames,
              int *bits_reg_mcc, int *bits_reg_mcc_uv,
              int *bits_reg_ovh, int *bits_reg_mode, int *bits_motion,
              int gen_bitstr_flag, int complexityMode)
{
  u_char
    *origY = origfr->yptr,
    *origU = origfr->uptr,
    *origV = origfr->vptr,
    *predY = predfr->yptr,
    *predU = predfr->uptr,
    *predV = predfr->vptr,
    *recoY = recofr->yptr,
    *recoU = recofr->uptr,
    *recoV = recofr->vptr;

  int
    lambda = (int)lambda_Y,
    x,y,k,l,m,p,q, tmp = 0, xPred, yPred, qPred,
    qp_UV, lambda_UV, 
    sxmin, sxmax, symax, symin, /* boudning points of the segment */
    sizex_ch, sizey_ch,          /* chroma frame size */
    reg_sx, reg_sy, /* region size */
    cnc_y[4] = {0}, cnc_u = 0, cnc_v = 0,
    class,            /* 8x8 Class of the prediction block*/
    codingMethod, dct4x4Meth,   /* Indicates selected transform */
    energy = 0,  /* sum of squares of pixel values in 8x8 blk */
    methCoeffs[64], dctCoeffs[64], 
    rotateFlag,
    PredBlk[64], OrigBlk[64], DiffBlk[64],
    PredBlk_u[64], PredBlk_v[64],
    decoder_bits, /* bits for the decoded block */
    bit_for_cbs,  /* bits spent by 4x4s */
    bit_for_dct,  /* bits spent by DCTs */
    bit_for_mes,  /* bits spent for method selection */ 
    ExtBlk[16], se_after, numCodedBlocks,
    pels, fastBest = 0, prevLagr = 5000000;


  /*
   * Initialize buffers for coding information.
   */
  if(gen_bitstr_flag) {
    triplet_counter = intra_dc_counter = doublet_counter = 0;
    vq_counter = rdc_counter = 0;
  }

  *bits_reg_mode = *bits_reg_ovh = *bits_reg_mcc = *bits_reg_mcc_uv = 0;

  sxmin = region->xmin;
  sxmax = region->xmax;
  symax = region->ymax;
  symin = region->ymin;
  sizex_ch = sizex >> 1;
  sizey_ch = sizey >> 1;

  reg_sx = reg_sy = regionSize >> 3;

/* qpBlockLuma used to pass qpY to vdxGetDctTable() and vdxGetKltTable() */
  qpBlockLuma = qp;

  qp_UV = annexTlookup[qp];
  lambda_UV = (int)(0.85 * qp_UV * qp_UV + 0.5);

  /*
   * Scan lumiance block(s).
   */
  for (y = symin, yPred = (blockInd & 2) << 2; y < symax; 
       y += 8, yPred += 8) {
    for (x = sxmin, xPred = (blockInd & 1) << 3; x < sxmax; 
         x += 8, xPred += 8) {

        /*
         * Copy blocks from frame.
         */
        for (k = 0; k < 8; k++) {
          p = k << 3;
          q = (y + k) * sizex + x;
          qPred = ((yPred + k) << 4) + xPred;

          for (l = 0; l < 8; l++, p++, q++, qPred++) {
            PredBlk[p] = (int)(predY[qPred]);
            OrigBlk[p] = (int)(origY[q]);
            DiffBlk[p] = OrigBlk[p] - PredBlk[p];
          }
        }

        /* 
         * Encode the block if energy of the difference block
         * is more than threshold. Threshold is obtained by 
         * calculating Lagrangian cost for minimum bits needed
         * for 8x8 coding and adding margin to that cost.
         */
        energy = calcBlockEnergy(DiffBlk, SIZE_8x8);

        if (energy >= MIN_BITS_8x8Y * lambda + MIN_IMPROVEMENT) {

          /*
           * Try to code the prediction error block.
           */
          codingMethod = enc8x8Block(DiffBlk, methCoeffs, qp, lambda, 
                                     AMOUNT_OF_8x8_METH + 1,
                                     0, &bit_for_cbs, &bit_for_dct, 
                                     &bit_for_mes, &se_after, PredBlk,
                                     dctCoeffs, LUMA, INTER_BLOCK, 0,
                                     fastBest, prevLagr, 0, 1, complexityMode);

          /*
           * Check if block is actually coded.
           */
          if (codingMethod > 0) {

            cnc_y[tmp] = 1;
            *bits_reg_mcc += bit_for_cbs + bit_for_dct + bit_for_mes;

            /*
             * Reconstruct the 8x8 prediction error block.
             */
            decoder_bits = dec8x8(PredBlk, DiffBlk, methCoeffs, dctCoeffs, qp,
                                  codingMethod, 0, LUMA, INTER_BLOCK, 0);

            /*
             * Store the coding information for bitstream dumping.
             */
            if (gen_bitstr_flag) {

              /*
               * Store 8x8 coding method.
               */
              class = changeClassNew(0, 0, qp, LUMA, INTER_BLOCK, 
                                     &rotateFlag);

              store_8x8_info1(codingMethod, class, dctCoeffs, INTER_BLOCK);

              /*
               * If 8x8 method signals that 4x4 Quadrants are coded separately,
               * store 4x4 coding method information and coefficient 
               * information for DCT and KLT coded 4x4 blocks.
               */
              if (codingMethod == METH_4x4_15) {
                store_8x8_info2(methCoeffs, codingMethod);
                dct4x4Meth = findDctMeth(methCoeffs, codingMethod, NULL);
                if (dct4x4Meth >= METH_4x4_DCT_1)
                  store_8x8_info3(dctCoeffs, dct4x4Meth, LUMA, INTER_BLOCK);
              }

              /*
               * If 8x8 block is coded with 8x8 DCT, 8x4 DCT or KLT, store 
               * transform coefficient information.
               */
              else if ((codingMethod >= METH_8x8_DCT &&
                        codingMethod <= METH_8x4_DCT_4) ||
                       (codingMethod >= METH_KLT_8x8_N_0_8 &&
                        codingMethod <= METH_KLT_8x8_N_7_8) ||
                       codingMethod == METH_DCT_8x8_N_0_8 ||
                       codingMethod == METH_DCT_8x8_N_4_8)
                store_8x8_info3(methCoeffs, codingMethod, LUMA, INTER_BLOCK);
            }
          }
        } 

        pedRecoInterBlk((int32*)PredBlk, (int32*)DiffBlk, recoY, y, x, sizex, 8,
                        cnc_y[((y - symin)>>3) * reg_sx + ((x - sxmin)>>3)]);
        tmp++;
    }
  }

  /*
   * Store the buffer counters for luminance information.
   */
  if (gen_bitstr_flag) {
    triplet_counter_Y = triplet_counter;
    intra_dc_counter_Y = intra_dc_counter;
    doublet_counter_Y = doublet_counter;
    vq_counter_Y = vq_counter;
    rdc_counter_Y = rdc_counter;
  }

  /* 
   * Code chrominance blocks.
   */
  xPred = (blockInd & 1) << 2;
  yPred = (blockInd & 2) << 1;
  x = sxmin >> 1;
  y = symin >> 1;

  switch (regionSize) {
    
    /*
     * 8x8 chrominance block of 16x16 macroblock.
     */
    case 16:

      /*
       * Copy U blocks from frame.
       */
      for (k = 0; k < 8; k++) {
        p = k << 3;
        q = (y + k) * sizex_ch + x;

        for (l = 0; l < 8; l++, p++, q++) {
          PredBlk_u[p] = (int)(predU[p]);
          OrigBlk[p] = (int)(origU[q]);
          DiffBlk[p] = OrigBlk[p] - PredBlk_u[p];
        }
      }

      /* 
       * Encode the block if energy of the difference block
       * is more than a threshold.
       */
      energy = calcBlockEnergy(DiffBlk, 64);
      if (energy >= MIN_BITS_8x8C * lambda_UV + MIN_IMPROVEMENT) {

        /*
         * Try to code the prediction error block.
         */
        codingMethod = enc8x8Block(DiffBlk, methCoeffs, qp_UV, lambda_UV,
                                   AMOUNT_OF_8x8_METH + 1, 0, &bit_for_cbs, 
                                   &bit_for_dct, &bit_for_mes, &se_after,
                                   PredBlk_u, dctCoeffs, CHROMA, INTER_BLOCK,
                                   0, fastBest, prevLagr, 0, 1, complexityMode);

        /*
         * Check if block is actually coded.
         */
        if (codingMethod > 0) {

          cnc_u = 1;
          *bits_reg_mcc_uv += bit_for_cbs + bit_for_dct + bit_for_mes;

          /*
           * Reconstruct the 8x8 prediction error block.
           */
          decoder_bits = dec8x8(PredBlk_u, DiffBlk, methCoeffs, dctCoeffs,
                                qp_UV, codingMethod, 0, CHROMA,
                                INTER_BLOCK, 0);

          if (gen_bitstr_flag) {

            /*
             * Store 8x8 coding method.
             */
            class = changeClassNew(0, 0, qp_UV, CHROMA, INTER_BLOCK, 
                                   &rotateFlag);

            store_8x8_info1(codingMethod, class, dctCoeffs, INTER_BLOCK);

            /*
             * If 8x8 method signals that 4x4 Quadrants are coded separately,
             * store 4x4 coding method information and coefficient 
             * information for DCT and KLT coded 4x4 blocks.
             */
            if (codingMethod == METH_4x4_15) {
              store_8x8_info2(methCoeffs, codingMethod);

              dct4x4Meth = findDctMeth(methCoeffs, codingMethod, NULL);
              if (dct4x4Meth >= METH_4x4_DCT_1)
                store_8x8_info3(dctCoeffs, dct4x4Meth, CHROMA, INTER_BLOCK);
            }

            /*
             * If 8x8 block is coded with 8x8 DCT, 8x4 DCT or KLT, store 
             * transform coefficient information.
             */
            else if ((codingMethod >= METH_8x8_DCT &&
                      codingMethod <= METH_8x4_DCT_4) ||
                     (codingMethod >= METH_KLT_8x8_N_0_8 &&
                      codingMethod <= METH_KLT_8x8_N_7_8) ||
                     codingMethod >= METH_DCT_8x8_N_0_8 ||
                     codingMethod <= METH_DCT_8x8_N_4_8)
              store_8x8_info3(methCoeffs, codingMethod, CHROMA, INTER_BLOCK);
          }
        }
      }

      /*
       * Create final reconstruction block.
       */
      pedRecoInterBlk((int32*)PredBlk_u, (int32*)DiffBlk, recoU, y, x, 
                      sizex_ch, 8, cnc_u);

      /*
       * Copy V blocks from frame.
       */
      for (k = 0; k < 8; k++) {
        p = k << 3;
        q = (y + k) * sizex_ch + x;

        for (l = 0; l < 8; l++, p++, q++) {
          PredBlk_v[p] = (int)(predV[p]);
          OrigBlk[p] = (int)(origV[q]);
          DiffBlk[p] = OrigBlk[p] - PredBlk_v[p];
        }
      }

      /* 
       * Encode the block if energy of the difference block
       * is more than a threshold.
       */
      energy = calcBlockEnergy(DiffBlk, 64);
      if (energy >= MIN_BITS_8x8C * lambda_UV + MIN_IMPROVEMENT) {

        /*
         * Try to code the prediction error block.
         */
        codingMethod = enc8x8Block(DiffBlk, methCoeffs, qp_UV, lambda_UV,
                                   AMOUNT_OF_8x8_METH + 1, 0, &bit_for_cbs, 
                                   &bit_for_dct, &bit_for_mes, &se_after,
                                   PredBlk_v, dctCoeffs, CHROMA, INTER_BLOCK,
                                   0, fastBest, prevLagr, 0, 1, complexityMode);

        /*
         * Check if block is actually coded.
         */
        if (codingMethod > 0) {

          cnc_v = 1;
          *bits_reg_mcc_uv += bit_for_cbs + bit_for_dct + bit_for_mes;

          /*
           * Reconstruct the 8x8 prediction error block.
           */
          decoder_bits = dec8x8(PredBlk_v, DiffBlk, methCoeffs, dctCoeffs,
                                qp_UV, codingMethod, 0, CHROMA,
                                INTER_BLOCK, 0);

          if (gen_bitstr_flag) {

            /*
             * Store 8x8 coding method.
             */
            class = changeClassNew(0, 0, qp_UV, CHROMA, INTER_BLOCK, 
                                   &rotateFlag);

            store_8x8_info1(codingMethod, class, dctCoeffs, INTER_BLOCK);

            /*
             * If 8x8 method signals that 4x4 Quadrants are coded separately,
             * store 4x4 coding method information and coefficient 
             * information for DCT and KLT coded 4x4 blocks.
             */
            if (codingMethod == METH_4x4_15) {
              store_8x8_info2(methCoeffs, codingMethod);

              dct4x4Meth = findDctMeth(methCoeffs, codingMethod, NULL);
              if (dct4x4Meth >= METH_4x4_DCT_1)
                store_8x8_info3(dctCoeffs, dct4x4Meth, CHROMA, INTER_BLOCK);
            }

            /*
             * If 8x8 block is coded with 8x8 DCT, 8x4 DCT or KLT, store 
             * transform coefficient information.
             */
            else if ((codingMethod >= METH_8x8_DCT &&
                      codingMethod <= METH_8x4_DCT_4) ||
                     (codingMethod >= METH_KLT_8x8_N_0_8 &&
                      codingMethod <= METH_KLT_8x8_N_7_8) ||
                     codingMethod == METH_DCT_8x8_N_0_8 ||
                     codingMethod == METH_DCT_8x8_N_4_8)
              store_8x8_info3(methCoeffs, codingMethod, CHROMA, INTER_BLOCK);
          }
        }
      }

      /*
       * Create final reconstruction block.
       */
      pedRecoInterBlk((int32*)PredBlk_v, (int32*)DiffBlk, recoV, y, x, 
                      sizex_ch, 8, cnc_v);
      break;

    /*
     * 4x4 chrominance block of splited macroblock.
     */
    case 8:

      /*
       * Copy U blocks from frame.
       */
      for (k = 0; k < 4; k++) {
        p = k << 2;
        q = (y + k) * sizex_ch + x;
        qPred = ((yPred + k) << 3) + xPred;

        for (l = 0; l < 4; l++, p++, q++, qPred++) {
          PredBlk_u[p] = (int)(predU[qPred]);
          OrigBlk[p] = (int)(origU[q]);
          DiffBlk[p] = OrigBlk[p] - PredBlk_u[p];
        }
      }

      /* 
       * Encode the block if energy of the difference block
       * is more than a threshold.
       */
      energy = calcBlockEnergy(DiffBlk, 16);
      if (energy >= MIN_BITS_4x4C * lambda_UV + MIN_IMPROVEMENT) {

        /*
         * Extract the reference pixels for extrapolation and try
         * to code the block.
         */
        pels = getExtSouJunct(PredBlk_u, ExtBlk, COLS_4x4, ROWS_4x4);
        bit_for_cbs = enc4x4Block(DiffBlk, PredBlk_u, OrigBlk, &se_after, 
                                  &bit_for_dct, methCoeffs, 0, lambda_UV, 
                                  qp_UV, dctCoeffs, ExtBlk, pels, CHROMA,
                                  INTER_BLOCK, energy, SIZE_4x4, fastBest,
                                  0, 0, energy,1,complexityMode);

        /*
         * Check if block is actually coded.
         */
        if (bit_for_cbs != -1) { 

          cnc_u = 1;
          *bits_reg_mcc_uv += bit_for_cbs + bit_for_dct;
          codingMethod = METH_4x4_1;
          for (m=16; m<64; m++)
            dctCoeffs[m] = 0;

          /*
           * Reconstruct the 4x4 prediction error block.
           */
          decoder_bits = dec4x4(PredBlk_u, DiffBlk, methCoeffs, dctCoeffs,
                                qp_UV, ExtBlk, pels, CHROMA,
                                INTER_BLOCK, SIZE_4x4);

          if (gen_bitstr_flag) {

            /*
             * Store 4x4 coding method information and coefficient 
             * information if 4x4 block is DCT or KLT coded.
             */
            store_8x8_info2(methCoeffs, codingMethod);
            dct4x4Meth = findDctMeth(methCoeffs, codingMethod, NULL);
            if (dct4x4Meth >= METH_4x4_DCT_1)
              store_8x8_info3(dctCoeffs, dct4x4Meth, CHROMA, INTER_BLOCK);
          }
        }
      }

      /*
       * Create final reconstruction block.
       */
      pedRecoInterBlk((int32*)PredBlk_u, (int32*)DiffBlk, recoU, y, x,
                      sizex_ch, 4, cnc_u);

      /*
       * Copy V blocks from frame.
       */
      for (k = 0; k < 4; k++) {
        p = k << 2;
        q = (y + k) * sizex_ch + x;
        qPred = ((yPred + k) << 3) + xPred;

        for (l = 0; l < 4; l++, p++, q++, qPred++) {
          PredBlk_v[p] = (int)(predV[qPred]);
          OrigBlk[p] = (int)(origV[q]);
          DiffBlk[p] = OrigBlk[p] - PredBlk_v[p];
        }
      }

      /* 
       * Encode the block if energy of the difference block
       * is more than a threshold.
       */
      energy = calcBlockEnergy(DiffBlk, 16);
      if (energy >= MIN_BITS_4x4C * lambda_UV + MIN_IMPROVEMENT) {

        /*
         * Extract the reference pixels for extrapolation and try
         * to code the block.
         */
        pels = getExtSouJunct(PredBlk_v, ExtBlk, COLS_4x4, ROWS_4x4);
        bit_for_cbs = enc4x4Block(DiffBlk, PredBlk_v, OrigBlk, &se_after, 
                                  &bit_for_dct, methCoeffs, 0, lambda_UV, 
                                  qp_UV, dctCoeffs, ExtBlk, pels, CHROMA,
                                  INTER_BLOCK, energy, SIZE_4x4, fastBest,
                                  0, 0, energy,1,complexityMode);

        /*
         * Check if block is actually coded.
         */
        if (bit_for_cbs != -1) { 

          cnc_v = 1;
          *bits_reg_mcc_uv += bit_for_cbs + bit_for_dct;
          codingMethod = METH_4x4_1;
          for (m=16; m<64; m++)
            dctCoeffs[m] = 0;

          /*
           * Reconstruct the 4x4 prediction error block.
           */
          decoder_bits = dec4x4(PredBlk_v, DiffBlk, methCoeffs, dctCoeffs,
                                qp_UV, ExtBlk, pels, CHROMA,
                                INTER_BLOCK, SIZE_4x4);

          if (gen_bitstr_flag) {

            /*
             * Store 4x4 coding method information and coefficient 
             * information if 4x4 block is DCT or KLT coded.
             */
            store_8x8_info2(methCoeffs, codingMethod);
            dct4x4Meth = findDctMeth(methCoeffs, codingMethod, NULL);
            if (dct4x4Meth >= METH_4x4_DCT_1)
              store_8x8_info3(dctCoeffs, dct4x4Meth, CHROMA, INTER_BLOCK);
          }
        }
      }

      /*
       * Create final reconstruction block.
       */
      pedRecoInterBlk((int32*)PredBlk_v, (int32*)DiffBlk, recoV, y, x,
                      sizex_ch, 4, cnc_v);

      break;
  }

  /*
   * Send coding mode.
   */
  if (regionSize == 16)
    numCodedBlocks = cnc_y[0] + cnc_y[1] + cnc_y[2] + cnc_y[3] +
                     cnc_u + cnc_v;
  else
    numCodedBlocks = cnc_y[0] + cnc_u + cnc_v;

  *pecMode = (numCodedBlocks > 0) ? PECMODE_CODED : PECMODE_SKIPPED;

  *bits_reg_mode = SendRegionMode(regionMode, *pecMode, predPossible,
                                  regionSize, gen_bitstr_flag);

  /* Send reference frame information if long-term memory is used.
   * For a splitted macroblock this was done when sending split
   * information
   */
  if(regionSize == 16 && numActiveRefFrames > 1){
    (*bits_reg_mode) += dumpLongtermRVLC(region->refFrameNumber, gen_bitstr_flag);
  }

  /* 
   * Send prediction information for motion if needed.
   */
  if (predPossible &&
      (regionMode == REGMODE_INTER_PRED ||
       regionMode == REGMODE_INTER_PRED_DELTA))
    *bits_motion = SendMotionPredInfo(predDirsAvail, predDir,
                                      predSelNeeded, predSel, gen_bitstr_flag);
  else
    *bits_motion = 0;

  /* 
   * Send motion information if needed.
   */
  if (regionMode == REGMODE_INTER_NOPRED)
    *bits_motion += dumpAffineMotionParams(motCoefs, aQP, gen_bitstr_flag);
  else if (regionMode == REGMODE_INTER_PRED_DELTA)
    *bits_motion += dumpAffineMotionParams(motCoefs, aQP, gen_bitstr_flag);

  if (*pecMode == PECMODE_CODED){

    /* 
     * Send CNC information.
     */
    *bits_reg_ovh = SendCncInfoDct(cnc_y, cnc_u, cnc_v, regionMode,
                                   regionSize);

    if (gen_bitstr_flag) {

      /*
       * Send prediction error coding information for
       * luminance.
       */
      tmp = dumpYInfo(cnc_y, sxmin, sxmax, symin, symax, qp);
      if (tmp != *bits_reg_mcc) {
        fprintf(stderr, "MCC y Inter (send) %d != %d (calc)\n",
                         tmp, *bits_reg_mcc);
        fprintf(stdout, "MCC y Inter (send) %d != %d (calc)\n",
                         tmp, *bits_reg_mcc);
        exit(-1);
      }

      /*
       * Send prediction error coding information for
       * luminance.
       */
      tmp = dumpUvInfo(&cnc_u, &cnc_v, regionSize, qp_UV);
      if (tmp != *bits_reg_mcc_uv) {
        fprintf(stderr ,"MCC uv Inter %d != %d\n", tmp,
                        *bits_reg_mcc_uv);
        fprintf(stdout ,"MCC uv Inter %d != %d\n", tmp,
                        *bits_reg_mcc_uv);
        exit(-1);
      }
    }

  }
  else {
    *bits_reg_ovh = 0;

    if (*bits_reg_mcc != 0 || *bits_reg_mcc_uv != 0) {
      fprintf(stderr, "Nonzero counters for MCC\n");
      fprintf(stdout, "Nonzero counters for MCC\n");
      exit(-1);
    }
  }

  return 0;
}

static 
void SendMacroblockYUV(struct frame *origfr, struct frame *predfr,
                       struct frame *recofr, int qp, int aQP,
                       object *object_list, int indMBlk,
                       int sizex, int sizey, float lambda,
                       int16 *slicePredMap, int currSliceID,
                       int *sliceXmin, int *sliceXmax,
                       int *sliceYmin, int *sliceYmax,
                       int numActiveRefFrames,
                       int *bitsTotalDct, int *bitsTotalDctC,
                       int *bitsTotalOvh,
                       int *bitsTotalMode, int *bitsTotalMotion,
                       int gen_bitstr_flag, int complexityMode)
{
  object
    *currRegion;

  int
    indRegion, indFirst, indLast, indPred,
    size_fr = sizex * sizey,
    bitsMode, bitsMotion, bitsOvh, bitsOvhC, bitsDct, bitsDctC,
    predSelNeeded, regionSize,
    tmp, lbl = -(currSliceID+1);

  *bitsTotalDct    = 0;
  *bitsTotalDctC   = 0;
  *bitsTotalOvh    = 0;

  *bitsTotalMode = *bitsTotalMotion = 0;

  if (gen_bitstr_flag)
    alloc_triplets_if_needed(size_fr);

  currRegion = &object_list[indMBlk];

  /* Send split information for macroblock */
  if(currRegion->regionMode == REGMODE_SPLITTED){

    *bitsTotalMode +=
      SendRegionMode(REGMODE_SPLITTED, PECMODE_SKIPPED,
                     currRegion->predPossible, 16, gen_bitstr_flag);

    /* Send reference frame information if long-term memory is used */
    if(numActiveRefFrames > 1){
      *bitsTotalMode += dumpLongtermRVLC(currRegion->refFrameNumber,
                                         gen_bitstr_flag);
    }

    /* Four 8x8 blocks to be sent */
    indFirst = indMBlk+1;
    indLast  = indMBlk+4;
  }
  else{

    /* One 16x16 macroblocks to be sent */
    indFirst = indMBlk;
    indLast  = indMBlk;
  }

  /* Region parameters for whole macroblock (1 or 4 regions) */
  for (indRegion = indFirst, indPred = 0; indRegion <= indLast; indRegion++,
       indPred++) {

    currRegion = &object_list[indRegion];
    regionSize = (currRegion->no_of_pels == 64) ? 8 : 16;

    tmp = (currRegion->ymin >> 3) * (sizex >> 3) + (currRegion->xmin >> 3);

    if(currRegion->regionMode == REGMODE_INTRA){

      /* only to be consistent with dumping of statistics */
      currRegion->pecMode = 0;

      if (regionSize == 16)
        slicePredMap[tmp] =
        slicePredMap[tmp + 1] =
        slicePredMap[tmp + (sizex >> 3)] =
        slicePredMap[tmp + (sizex >> 3) + 1] = lbl;
      else
        slicePredMap[tmp] = lbl;

      /* Update the geometry of the intra part of
         slice which developped until now */

      if (currRegion->ymin < *sliceYmin)
        *sliceYmin = currRegion->ymin;
      if (currRegion->xmin < *sliceXmin)
        *sliceXmin = currRegion->xmin;
      if (currRegion->ymin+regionSize-1 > *sliceYmax)
        *sliceYmax = currRegion->ymin+regionSize-1;
      if (currRegion->xmin+regionSize-1 > *sliceXmax)
        *sliceXmax = currRegion->xmin+regionSize-1;

      SendIntraMccRegion(origfr, recofr, slicePredMap, lbl,
                         *sliceXmin, *sliceXmax, *sliceYmin, *sliceYmax,
                         currRegion, regionSize, currRegion->predPossible,
                         sizex, sizey, INTRA_BLOCK, qp, lambda,
                         &bitsDct, &bitsDctC, &bitsOvh, &bitsOvhC,
                         &bitsMode, gen_bitstr_flag, complexityMode);

      *bitsTotalDct  += bitsDct;
      *bitsTotalDctC += bitsDctC;
      *bitsTotalOvh  += bitsOvh + bitsOvhC;
      *bitsTotalMode += bitsMode;
    }

    else if(currRegion->regionMode == REGMODE_INTER_NOPRED     ||
            currRegion->regionMode == REGMODE_INTER_PRED       ||
            currRegion->regionMode == REGMODE_INTER_PRED_DELTA ||
            currRegion->regionMode == REGMODE_INTER_COPY){

      if (regionSize == 16)
        slicePredMap[tmp] =
        slicePredMap[tmp + 1] =
        slicePredMap[tmp + (sizex >> 3)] =
        slicePredMap[tmp + (sizex >> 3) + 1] = currSliceID;
      else
        slicePredMap[tmp] = currSliceID;

      if(currRegion->regionMode == REGMODE_INTER_PRED_DELTA ||
         currRegion->regionMode == REGMODE_INTER_PRED)
        if(currRegion->predDir == PREDDIR_UP)
          predSelNeeded = currRegion->predSelNeededUp;
        else
          predSelNeeded = currRegion->predSelNeededLeft;
      else
        predSelNeeded = PREDSEL_NOTNEEDED;

      SendMccRegion(origfr, predfr, indPred, recofr, currRegion,
                    sizex, sizey, regionSize,
                    currRegion->regionMode,    currRegion->predPossible,
                    currRegion->predDirsAvail, currRegion->predDir,
                    predSelNeeded, currRegion->predSel,
                    currRegion->motCoefOrth, &currRegion->pecMode,
                    qp, aQP, lambda, numActiveRefFrames,
                    &bitsDct, &bitsDctC, &bitsOvh, &bitsMode, &bitsMotion,
                    gen_bitstr_flag, complexityMode);

      *bitsTotalDct    += bitsDct;
      *bitsTotalDctC   += bitsDctC;
      *bitsTotalOvh    += bitsOvh;
      *bitsTotalMode   += bitsMode;
      *bitsTotalMotion += bitsMotion;

    }

    else{
      fprintf(stderr,
              "Unknown region mode %d in SendMacroblockYUV\n",
              currRegion->regionMode);
      exit(11);
    }

    /* 16x16 region */
    if(currRegion->no_of_pels == 256 &&
       currRegion->regionMode == REGMODE_SPLITTED)
        currRegion->pecMode = 0;
  }
}

void SendMacroblockAccumBits(struct frame *currFr, struct frame *predBuffer,
                             struct frame *recoFr, int qp, int aQP,
                             object *objectList, int indMBlk,
                             int picSizeX, int picSizeY, float lambda,
                             int16 *slicePredMap, int currSliceID,
                             int *sliceXmin, int *sliceXmax,
                             int *sliceYmin, int *sliceYmax,
                             int numActiveRefFrames,
                             Bits *bits, int complexityMode)
{
  int
    bitsMode, bitsMotion,
    bitsCoefY, bitsCoefC, bitsOvh;

  /* Code the prediction error and dump the bitstream */
  SendMacroblockYUV(currFr, predBuffer, recoFr, qp, aQP, objectList, indMBlk,
                    picSizeX, picSizeY, lambda, slicePredMap, currSliceID,
                    sliceXmin, sliceXmax, sliceYmin, sliceYmax,
                    numActiveRefFrames,
                    &bitsCoefY, &bitsCoefC, &bitsOvh,
                    &bitsMode, &bitsMotion, 1, complexityMode);

  bits->coefy   += bitsCoefY;
  bits->coefu   += bitsCoefC;
  bits->cncy    += bitsOvh;
  bits->mode    += bitsMode;
  bits->motcoef += bitsMotion;

}
