#ifndef _PED_H_
#define _PED_H_

#include "nrctypes.h"
#include "mainFlags.h"
#include "nccglob.h"
#include "decDemux.h"
void pedRecoInterBlk(int32 *predFrame, int32 *diffFrame, u_char *recoFrame, 
                     int32 y, int32 x, int32 xRes, int32 delta, 
                     int32 codedFlag);

void pedRecoCopyRegionYUV(struct frame *predFrame, struct frame *recoFrame, 
                          object *region, int32 picSizeX, int32 blockInd);

int32 pedLoadMethod(int32 classAfterRotations, int32 *codingMethod,
                    int32 chroma, int32 blkSize,
                    vdxInstance_t *instance, int* BitError);
int32 pedLoad4x4Info(int32 *methCoeffs, int32 chroma, int32 mode,
                     int32 qp, int32 blkSize_dummy, int32 codingMethod,
                     vdxInstance_t *instance, int* BitError);
int32 pedLoadDctCoeffs(int32 *methCoeffs, int32 codingMethod, int32 chroma,
                       int32 qp, int32 mode, int *save4,
                       vdxInstance_t *instance, int* BitError);


/************************************************************************
 *  DecodePFrame() performs the complete prediction & reconstruction
 *  of one frame (recfr) from the previous reconstructed frame (prevfr).
 *  It returns the total number of bits decoded.
 *
 *  Input args:
 *    prevfr (a pointer to Yuv frame buffer holding the reference frame)
 *    pic_sizex, pic_sizey (luminance frame size)
 *    fr_no (the frame number of the currently decoded frame)
 *    qp (quantizer parameter for DCT in prediction error coding)
 * 
 *  Output args:
 *    predfr (a pointer to Yuv frame buffer holding the mc predicted frame)
 *    recfr (a pointer to Yuv frame buffer holding the reconstructed frame)
 *    object_list (array of objects of the decoded regions)
 *    no_of_regions (total number of regions in the decoded frame)
 *    mc_coeff_bits (bits to encode the motion parameters)
 *    merge_bits (number of bits to encode the MERGE bit sequence)
 *    split_bits (number of bits to encode the SPLIT bit sequence)
 *    dct_y_bits (number of bits to encode the DCT coeffs of Y prediction error)
 *    dct_u_bits (number of bits to encode the DCT coeffs of u prediction error)
 *    dct_v_bits (number of bits to encode the DCT coeffs of v prediction error)
 *    ovh_y_bits (number of overhead bits in Y prediction error coding)
 *    ovh_uv_bits (number of overhead bits in u and v prediction error coding)
 *    region_mode_bits (number of bits to encode the modes of the regions)
 *
 *  Warnings:
 *    object_list needs to be pre-allocated large enough.
 *    prevfr, predfr, recfr need to be pre-allocated for Yuv frames.
 *************************************************************************/

int32 DecodePFrame(struct frame **longTermMem, struct frame *predfr,
                   struct frame *recfr, int picSizeX, int picSizeY, int l,
                   objectBuffer_t *object_list, 
                   int16 *slicePredMap,
                   int sliceMode, int qp, int aQP, int numActiveRefFrames,
                   int *orderedRefFrames,
                   int32 *bitsTotalMotion,
                   int32 *bitsTotalDctY, int32 *bitsTotalDctU, 
                   int32 *bitsTotalDctV, int32 *bitsTotalOvrY,
                   int32 *bitsTotalOvrC, 
                   int32 *bitsTotalMode,
                   int32 *cumBitsSliceHeaders,
                   struct frame *outfr, 
                   vdxInstance_t *instance, 
                   int *BitError,
                   int useLoopFilter);
#endif
