/*
 * encMux.c
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Contains:
 *    Functions for video multiplexer.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "nrctypes.h"
#include "nccglob.h"
#include "encMux.h"
#include "debug.h"
#include "mainFlags.h"
#include "tools.h"
#include "mcp.h"
#include "toolMCC.h"

#include "encMux.dat"

extern doublet *region_doublets;
extern triplet *vq_triplets;
extern triplet *region_triplets;

extern int *rdc_value;
extern int longTermMemSize;
extern int *region_intra_dc;
extern int triplet_counter;
extern int intra_dc_counter;
extern int
  triplet_counter,
  intra_dc_counter,
  doublet_counter,
  vq_counter,
  rdc_counter;
extern int
  triplet_counter_Y,
  intra_dc_counter_Y,
  doublet_counter_Y,
  vq_counter_Y,
  rdc_counter_Y;

extern u_char
  overheadNumbers4x4[AMOUNT_OF_4x4_OVH],
  overheadNumbers8x8[AMOUNT_OF_8x8_OVH];

static volatile int all_coding_bits, emulPreventionBits = 0, paddingBits = 3;

extern int
  tableInter[65],
  tableIntra[65];

static int dump8x8MethSel(int class, int method, int chroma, int mode,
                          int gen_bitstr_flag);

static int dump4x4MethSel(int class, int method, int chroma, int mode,
                          int gen_bitstr_flag);

int dump8x8BlockAcInfo(int mode, int chroma, int qp, int *tmpDouCnt,
                       int *tmpRdcCnt, int *tmpVqCnt, int *tmpTriCnt,
                       int gen_bitstr_flag);
int dump4x4BlockAcInfo(int mode, int chroma, int qp, int *tmpRdcCnt,
                       int *tmpVqCnt, int *tmpTriCnt, int gen_bitstr_flag);
static int SendMccTriplet(int last, int run, int level,
                          vlc_item *vlcTable, int vlcInd,
                          int levelBits, int runBits, int escapeInd,
                          int gen_bitstr_flag);



/*
 * dumpLongtermRVLC(int index)
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Parameters:
 *     index : index of the reference frame
 * Function:    
 *      Signals the reference frame indexes to the decoder 
 *		in Long Term Memory Mode.
 * Returns:
 *      Number of bits dumped to the bitstream
 */

int dumpLongtermRVLC(int index, int send)
{
  static int      first_invo = 1;
  static vlc_item longtermRVLC[10];

  if(first_invo) {
    bitInitLongtermRVLC(longtermRVLC);
    first_invo = 0;
  }

  if(send)
    return put_vlc_item2(longtermRVLC[index], 0);
  else
    return longtermRVLC[index].len;
}

/*
 * SendCncInfoDct(int *cnc_y, int cnc_u, int cnc_v,
 *		                  int regionMode, int regionSize)
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Parameters:
 *     cnc_y : CNC info for Luminance. ( 1 value for 8x8 and 4 values for macroblock)
 *	   cnc_u : CNC info for Chrominance U 
 *     cnc_v : CNC info for Chrominance Y 
 *     regionMode : REGMODE_INTRA or REGMODE_INTER
 *     regionSize :  8 or 16 w.r.t. a quadrant or macroblock.
 *
 * Function:    
 *      Dumps the CNC information of a macroblock or a block to the bitstream.
 *		
 * Returns:
 *      Number of bits dumped to the bitstream
 */
   
int SendCncInfoDct(int *cnc_y, int cnc_u, int cnc_v,
		   int regionMode, int regionSize)
{
  int ddump, bits=0;

  if(regionSize == 16){
    /* ddump is a 6 bit value composed of 
        [ cnc_y[0] cnc_y[1] cnc_y[2] cnc_y[3] cnc_u cnc_v ] */
    ddump = 32*cnc_y[0] + 16*cnc_y[1] +
             8*cnc_y[2] +  4*cnc_y[3] +
	     2*cnc_u    +    cnc_v;

    if(regionMode == REGMODE_INTRA) {
      bits += put_vlc_item2(vlcEncCncIntra64[tableIntra[ddump]], 0);
	  /* check if ddump is a valid index or not */
      if (tableIntra[ddump] ==18) /* 18 is the escape index */
		bits += put_bits(6, ddump, 0, 1);
    }
    else {
	  bits += put_vlc_item2(vlcEncCncInter64[tableInter[ddump]], 0);
      /* check if ddump is a valid index or not */
	  if (tableInter[ddump] ==15) /* 15 is the escape index */
		bits += put_bits(6, ddump, 0, 1);
    }
  }
  else{
    /* ddump is a 3 bit value composed of  [ cnc_y[0]  cnc_u cnc_v ] */
    ddump = 4*cnc_y[0] + 2*cnc_u + cnc_v;
    if(regionMode == REGMODE_INTRA) {
      bits += put_vlc_item2(vlcEncCncIntra8[ddump], 0);
    }
    else {
      bits += put_vlc_item2(vlcEncCncInter8[ddump-1], 0);
    }
  }

  return bits;
}

/*
 * SendYInfoDct(int *cncY, int regionMode, int regionSize)
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Parameters:
 *     cncY  : CNC info for Luminance
 *     regionMode : REGMODE_INTRA or REGMODE_INTER
 *     regionSize :  8 or 16 w.r.t. a quadrant or macroblock.
 *
 * Function:    
 *      Dumps triplet (Run-Length-Last) information for luminace
 *		
 * Returns:
 *      Number of bits dumped to the bitstream
 */

int SendYInfoDct(int *cncY, int regionMode, int regionSize)
{
  int
    notLastSent, numBlocks, blkIndex,
    bits      = 0,
    tmpTriCnt = 0;

  numBlocks = (regionSize == 16) ? 4 : 1;
  /* numBlocks = 4 => macroblock having 4 8x8 blocks, 
      numBlocks = 1 => one 8x8 block*/
  for(blkIndex = 0;blkIndex < numBlocks;blkIndex++) {
    if(regionMode == REGMODE_INTRA) 
      /* Dump DC level and sign */
      bits += Send_DC_Pred(region_intra_dc[blkIndex], LUMA);
    if(cncY[blkIndex]){
      notLastSent = 1;
      while(notLastSent){
        /* send triplet info about DCT coefficients*/
	bits += Send_Event(region_triplets[tmpTriCnt].last,
			   region_triplets[tmpTriCnt].run,
			   region_triplets[tmpTriCnt].level);
	notLastSent = !region_triplets[tmpTriCnt++].last;
      }
    }
  }

  if(tmpTriCnt != triplet_counter_Y) {
    fprintf(stdout, "tmpTriCnt %d != %d triplet_counter_Y (Dct Region)\n",
            tmpTriCnt, triplet_counter_Y);
    fflush(stdout); nccExit(1);
  }

  return bits;
}

/*
 * endCInfoDct(int cncU, int cncV, int regionMode, int regionSize)
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Parameters:
 *     cncU  : CNC info for Chrominance U
 *     cncV  : CNC info for Chrominance V
 *     regionMode : REGMODE_INTRA or REGMODE_INTER
 *     regionSize :  8 or 16 w.r.t. a quadrant or macroblock.
 *
 * Function:    
 *      Dumps triplet (Run-Length-Last) information for Chrominance
 *		
 * Returns:
 *      Number of bits dumped to the bitstream
 */

int SendCInfoDct(int cncU, int cncV, int regionMode, int regionSize)
{
  int
    notLastSent,
    bits      = 0,
    tmpTriCnt = triplet_counter_Y;

  if(regionSize == 16){

    /* 8x8 U component */
    if(regionMode == REGMODE_INTRA)
      /* Dump DC level and sign */
      bits += Send_DC_Pred(region_intra_dc[4], CHROMA);

    if(cncU){
      notLastSent = 1;
      while(notLastSent){
      /* send triplet info about DCT coefficients*/
	bits += Send_Event(region_triplets[tmpTriCnt].last,
			   region_triplets[tmpTriCnt].run,
			   region_triplets[tmpTriCnt].level);
	notLastSent = !region_triplets[tmpTriCnt++].last;
      }
    }

    /* 8x8 V component */
    if(regionMode == REGMODE_INTRA)
      bits += Send_DC_Pred(region_intra_dc[5], CHROMA);

    if(cncV){
      notLastSent = 1;
      while(notLastSent){
        /* send triplet info about DCT coefficients*/
	bits += Send_Event(region_triplets[tmpTriCnt].last,
			   region_triplets[tmpTriCnt].run,
			   region_triplets[tmpTriCnt].level);
	notLastSent = !region_triplets[tmpTriCnt++].last;
      }
    }
  }
  else{

    /* 4x4 U component */
    if(regionMode == REGMODE_INTRA)
      /* Dump DC level and sign */
      bits += Send_DC_Pred(region_intra_dc[4], CHROMA);

    if(cncU){
      notLastSent = 1;
      while(notLastSent){
        /* send triplet info about DCT coefficients*/
	bits += Send_Event4x4Color(region_triplets[tmpTriCnt].last,
				   region_triplets[tmpTriCnt].run,
				   region_triplets[tmpTriCnt].level);
	notLastSent = !region_triplets[tmpTriCnt++].last;
      }
    }

    /* 4x4 V component */
    if(regionMode == REGMODE_INTRA)
      /* Dump DC level and sign */
      bits += Send_DC_Pred(region_intra_dc[5], CHROMA);

    if(cncV){
      notLastSent = 1;
      while(notLastSent){
        /* send triplet info about DCT coefficients*/
	bits += Send_Event4x4Color(region_triplets[tmpTriCnt].last,
				   region_triplets[tmpTriCnt].run,
				   region_triplets[tmpTriCnt].level);
	notLastSent = !region_triplets[tmpTriCnt++].last;
      }
    }
  }

  if(tmpTriCnt != triplet_counter) {
    fprintf(stdout, "tmpTriCnt %d != %d triplet_counter (Dct Region)\n",
            tmpTriCnt, triplet_counter);
    fflush(stdout); nccExit(1);
  }

  return bits;
}


/*
 * dumpAffineMotionParams(int *motCoefQuant, int aQP, int send)
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Parameters:
 *     motCoefQuant  : Quantized motion coefficients
 *     aQP  :  Affine QP which is a function of the frame QP 
 *     send : 0 => generated bits not dumped to the bitstream
 *              1 => generated bits dumped to the bitstream
 *
 * Function:    
 *      Dumps the affine motion parameter zero-not-zero pattern 
 *     and the motion parameters to the bitstream if send = 1.
 *		
 * Returns:
 *      Number of bits dumped to the bitstream
 */

int dumpAffineMotionParams(int *motCoefQuant, int aQP, int send)
{
  int
    ind, i, bits = 0, sign, motCoefInd[6];

  const int th1   = 21;             /* Threshold for the 1st Escape code   */
  const int th2   = 85;             /* Threshold for the 2nd Escape code   */
  const int LEN_1 = 6;              /* Number of bits following 1st Escape */
  const int LEN_2 = 10;             /* Number of bits following 2nd Escape */
  const int MAX_AMPL = 1108;        /* Maximum amplitude of affine params  */
  const int LEN_TAB = 22;           /* Length of the Huff table for amplit */

  /* Generate zero pattern information for motion coefficients*/
  ind = 0;
  for(i = 0; i < 6; i++)
    if((motCoefInd[i] = motCoefQuant[i]/aQP) != 0)
      ind += (1 << i);

  if (!ind) {
    fprintf(stdout,
	    "All coeffs are sent as zero in dumpAffineMotionParams!\n");
    fprintf(stderr,
	    "All coeffs are sent as zero in dumpAffineMotionParams!\n");
    exit(-11);
  }

  if(send){

    /* Dump zero pattern */
    bits += put_vlc_item2(affZeroPat[ind - 1], 0);

    /* Dump nonzero affine motion parameters */  
    for(i=0; i < 6; i++){

      if((ind = abs(motCoefInd[i])) != 0){

        sign = ((motCoefInd[i]<0) ? 1:0);

        if (ind < th1){
          bits += put_vlc_item2(affAmpl[ind-1], 0);
          bits += put_bits(1, sign, 0, 1);
        }

        /* First Escape + 6 bits */
        else if (ind < th2){
          bits += put_vlc_item2(affAmpl[LEN_TAB - 2], 0);
          bits += put_bits(LEN_1, ind-th1,0,1);
          bits += put_bits(1, sign, 0, 1);
        }

        /* Second Escape + 10 bits */
        else if (ind <= MAX_AMPL){
          bits += put_vlc_item2(affAmpl[LEN_TAB - 1], 0);
          bits += put_bits(LEN_2, ind-th2,0,1);
          bits += put_bits(1, sign, 0, 1);
        }
        else{
          fprintf(stderr,"Affine motion parameter out of range: %d > %d\n",
                  ind, MAX_AMPL);
          exit(1);
        }
      }
    }
  }
  else{

    /* Bits for zero pattern */
    bits += affZeroPat[ind - 1].len;

    /* Bits for nonzero affine motion parameters */  
    for(i=0; i < 6; i++){

      if((ind = abs(motCoefInd[i])) != 0){

        sign = ((motCoefInd[i]<0) ? 1:0);

        if (ind < th1)
          bits += affAmpl[ind-1].len + 1;

        /* First Escape + 6 bits */
        else if (ind < th2)
          bits += affAmpl[LEN_TAB - 2].len + LEN_1 + 1;

        /* Second Escape + 10 bits */
        else if (ind <= MAX_AMPL)
          bits += affAmpl[LEN_TAB - 1].len + LEN_2 + 1;
        else{
          fprintf(stderr,"Affine motion parameter out of range: %d > %d\n",
                  ind, MAX_AMPL);
          exit(1);
        }
      }
    }
  }

  return bits;
}

/*
 * Send_DC_Pred(int level, int chroma)
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Parameters:
 *     level      : DC level
 *     chroma  :  Chrominance or Luminance indicator
 * Function:    
 *     Dumps the DC level and the sign of it to the bitstream. 
 * Returns:
 *      Number of bits dumped to the bitstream
 */

int Send_DC_Pred(int level, int chroma)
{
  int bits=0, tmpLevel, sign;

  static int first_invo=1;
  static vlc_item DC_Luma[8];
  static vlc_item DC_Chro[8];

  if(first_invo) {
    bitInitDC_Luma(DC_Luma);
    bitInitDC_Chro(DC_Chro);
    first_invo = 0;
  }

  if(level >= 400) {
    if(chroma == CHROMA) {
      bits += put_vlc_item2(DC_Chro[7], 0);
    }
    else {
      bits += put_vlc_item2(DC_Luma[7], 0);
    }
    tmpLevel = level - 400;
    bits += put_bits(8, tmpLevel, 0, 1);
  }
  else if(level >= 100) {
    tmpLevel = level - 100;
    bits += put_bits(8, tmpLevel, 0, 1);
  }
  else {
    tmpLevel = abs(level);

    if(chroma == CHROMA) {
      if(tmpLevel < 1) {
        bits += put_vlc_item2(DC_Chro[0], 0);
      }
      else if(tmpLevel < 2) {
        bits += put_vlc_item2(DC_Chro[1], 0);
      }
      else if(tmpLevel < 4) {
        bits += put_vlc_item2(DC_Chro[2], 0);
        tmpLevel -= 2;
        bits += put_bits(1, tmpLevel, 0, 1);
      }
      else if(tmpLevel < 8) {
        bits += put_vlc_item2(DC_Chro[3], 0);
        tmpLevel -= 4;
        bits += put_bits(2, tmpLevel, 0, 1);
      }
      else if(tmpLevel < 16) {
        bits += put_vlc_item2(DC_Chro[4], 0);
        tmpLevel -= 8;
        bits += put_bits(3, tmpLevel, 0, 1);
      }
      else if(tmpLevel < 32) {
        bits += put_vlc_item2(DC_Chro[5], 0);
        tmpLevel -= 16;
        bits += put_bits(4, tmpLevel, 0, 1);
      }
      else {
        bits += put_vlc_item2(DC_Chro[6], 0);
        tmpLevel -= 32;
        bits += put_bits(5, tmpLevel, 0, 1);
      }
    }
    else {
      if(tmpLevel < 1) {
        bits += put_vlc_item2(DC_Luma[0], 0);
      }
      else if(tmpLevel < 2) {
        bits += put_vlc_item2(DC_Luma[1], 0);
      }
      else if(tmpLevel < 4) {
        bits += put_vlc_item2(DC_Luma[2], 0);
        tmpLevel -= 2;
        bits += put_bits(1, tmpLevel, 0, 1);
      }
      else if(tmpLevel < 8) {
        bits += put_vlc_item2(DC_Luma[3], 0);
        tmpLevel -= 4;
        bits += put_bits(2, tmpLevel, 0, 1);
      }
      else if(tmpLevel < 16) {
        bits += put_vlc_item2(DC_Luma[4], 0);
        tmpLevel -= 8;
        bits += put_bits(3, tmpLevel, 0, 1);
      }
      else if(tmpLevel < 32) {
        bits += put_vlc_item2(DC_Luma[5], 0);
        tmpLevel -= 16;
        bits += put_bits(4, tmpLevel, 0, 1);
      }
      else {
        bits += put_vlc_item2(DC_Luma[6], 0);
        tmpLevel -= 32;
        bits += put_bits(5, tmpLevel, 0, 1);
      }
    }

    if(level != 0) {
      sign = level > 0 ? 0:1;
      bits += put_bits(1, sign, 0, 1);
    }

  }
  return bits;
}

/*
 * Send_Event(int last, int run, int level)
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Function:    
 *    Dumps DCT coefficients related triplet 
 *    (Run-Length-Last) codeword to the bitstream.
 * Returns:
 *      Number of bits dumped to the bitstream
 */
int Send_Event(int last, int run, int level)
{ 
  static int first_invo=1;
  static vlc_item DCT_vlc[103];
  static int triple2index[LAST][RUN][LEVEL];
  int
    sign, index, bits=0;
  
  if(first_invo){
    init_enc_DCT_vlc(DCT_vlc);
    init_enc_triple2index(triple2index);
    first_invo=0;
  }
  
  sign = level>0 ? 0:1;
  level = abs(level);

  if(!level || level>128){
    fprintf(stderr,"Invalid level of AC\n");
    exit(1);
  }

  index=triple2index[last][run][level];
  if(index!=102){
    bits += put_vlc_item2(DCT_vlc[index], 0);
    bits += put_bits(1,sign,0,1);
  }
  else{
    bits += put_vlc_item2(DCT_vlc[102], 0); /*escape*/
    bits += put_bits(1,last,0,1);
    bits += put_bits(6,run,0,1);
    bits += put_bits(1,sign,0,1);
    if(sign)
      level = 128-level; /* for negative */
    bits += put_bits(7,level,0,1);
  }
  return bits;
}

/*
 * Send_EventIPred(int last, int run, int level, int long_escape)
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Function:    
 *    Dumps Intra Prediction triplet (Run-Length-Last) codeword to the bitstream
 * Returns:
 *      Number of bits dumped to the bitstream
 */
int Send_EventIPred(int last, int run, int level, int long_escape)
{ 
  static int first_invo=1;
  static vlc_item DCT_vlc[103];
  static int triple2index[LAST][RUN][LEVEL];
  int
    sign, index, bits=0;
  
  if(first_invo){
    init_enc_DCTIPred_vlc(DCT_vlc);
    init_enc_triple2index_IPred(triple2index);
    first_invo=0;
  }

  sign = level>0 ? 0:1;
  level = abs(level);

  if(!level || level>2048){
    fprintf(stderr,"Invalid level of AC\n");
    exit(1);
  }

  if(long_escape && (level>127)) index=102;
  else  index=triple2index[last][run][level];
  if(index!=102){
    bits += put_vlc_item2(DCT_vlc[index], 0);
    bits += put_bits(1,sign,0,1);
  }
  else{
    bits += put_vlc_item2(DCT_vlc[102], 0); /*escape*/
    bits += put_bits(1,last,0,1);
    bits += put_bits(6,run,0,1);

    if(long_escape && level>127){
      bits += put_bits(8,128,0,1); /* extended escape */
      bits += put_bits(1,sign,0,1);
      if(sign)
	level = 2048 - level;
      bits += put_bits(11,level,0,1);
    }
    else{
      bits += put_bits(1,sign,0,1);
      if(sign)
	level = 128-level; /* for negative */
      bits += put_bits(7,level,0,1);
    }
  }
  return bits;
}

/*
 * Send_Event4x4Color(int last, int run, int level)
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Function:    
 *    Dumps 4x4 U and V chrominance triplet (Run-Length-Last) information to the bitstream
 * Returns:
 *      Number of bits dumped to the bitstream
 */
int Send_Event4x4Color(int last, int run, int level)
{ 
  static int first_invo=1;
  static vlc_item dct4x4Color[83];
  static int triple2index4x4Color[LAST][RUN4x4][LEVEL];
  int
    sign, index, bits=0;
  
  if(!level){
    fprintf(stderr,"Zero level of AC\n");
    exit(1);
  }

  if(first_invo){
    init_enc_dct4x4Color(dct4x4Color);
    init_enc_triple2index4x4Color(triple2index4x4Color);
    first_invo=0;
  }

  sign = level>0 ? 0:1;
  level = abs(level);
  index=triple2index4x4Color[last][run][level];
  if(index!=82){
    bits += put_vlc_item2(dct4x4Color[index], 0);
    bits += put_bits(1,sign,0,1);
  }
  else{
    bits += put_vlc_item2(dct4x4Color[82], 0); /*escape*/
    bits += put_bits(1,last,0,1);
    bits += put_bits(4,run,0,1);
    bits += put_bits(1,sign,0,1);
    if(sign)
      level = 128-level; /* for negative */
    bits += put_bits(7,level,0,1);
  }
  return bits;
}

int Send_CBPY(int index)
{
  static int first_invo=1;
  static vlc_item CBPY[16];

  if(first_invo){
    init_enc_CBPY(CBPY);
    first_invo=0;
  }

  return put_vlc_item2(CBPY[index], 0);

}

int Send_CBPC(int index)
{
  static int first_invo=1;
  static vlc_item CBPC[4];

  if(first_invo){
    init_enc_CBPC(CBPC);
    first_invo=0;
  }

  if(index>3){
    fprintf(stderr,"not supported macroblockk type in Send_CBPC\n");
    exit(1);
  }
  else
    return put_vlc_item2(CBPC[index], 0);

}

int put_vlc_item2(vlc_item item, int flush)
{
  return put_bits(item.len, item.code, flush,1);
}

int get_all_bits()
{
  return all_coding_bits + emulPreventionBits + paddingBits;
}

int get_padding_bits()
{
  return paddingBits;
}

int get_emulprevent_bits()
{
  return emulPreventionBits;
}

void reset_all_bits()
{
  all_coding_bits       = 0;
  emulPreventionBits    = 0;
  paddingBits           = 0;
}

int put_bits(int clen, int code, int flush, int emulationPrev)
{
  static int 
    two_powers[MAX_VLC_LEN],
    curr_bit,
    firstInvo = 1;

  static unsigned char 
    PSCbyte1, PSCbyte2,
    put_byte;

  int
    i, len = clen;
  
#ifndef PSC_BitStuff_Enable
  emulationPrev = 0;
#endif

  if (firstInvo)
  {
    two_powers[0] = 1;
    for (i = 1; i < MAX_VLC_LEN; i++)
      two_powers[i] = 2 * two_powers[i - 1];

    put_byte = 0;

    /* 
     * Do not do emulation prevention for the first PSC.
     */
    PSCbyte1 = PSCbyte2 = 1; 
    curr_bit = 7;
    firstInvo = 0;
  }

  all_coding_bits += clen;

  while (len > 0) 
  {
    if (code & two_powers[len - 1]) 
      put_byte |= two_powers[curr_bit];

    /*
     * Check if there is two consecutive bytes of zeros and
     * start code emulation prevention is requested.
     */
    if (emulationPrev && PSCbyte1 == 0x00 && PSCbyte2 == 0x00)
    {
      /* 
       * Check for PSC emulation.
       *
       * After writing 00000000 00000000 01000, force extra '1' bit
       * in order to avoid PSC 00000000 00000000 010000. 
       */
      if (curr_bit == 3 && put_byte == 0x40)
      {
        curr_bit--;
        put_byte |= two_powers[curr_bit];
        emulPreventionBits++;
      }      
      /* 
       * Check for SSC emulation.
       *
       * After writing 00000000 00000000 1, force extra '0' bit
       * in order to avoid SSC 00000000 00000000 11. 
       */
      else if (curr_bit == 7 && put_byte == 0x80)
      {
        curr_bit--;
        emulPreventionBits++;	
      }
      /*
       * Check for EOS emulation.
       *
       * After writing 00000000 00000000 01111, force extra '0' bit
       * in order to avoid EOS 00000000 00000000 011111.
       */
      else if (curr_bit == 3 && put_byte == 0x78)
      {
        curr_bit--;
        emulPreventionBits++;
      }
    }
    curr_bit--;

    /*
     * If full byte is done, flush it to bitstream.
     */
    if (curr_bit < 0) 
    {
      fwrite(&put_byte, 1, 1, bitstream_bit);
      PSCbyte1 = PSCbyte2;
      PSCbyte2 = put_byte;
      put_byte = 0;
      curr_bit = 7;
    }
    len--;
  }
  /*
   * Flush the bits after the codeword if requested.
   */
  if(curr_bit != 7 && flush)
  {
    /*
     * Prevent PSC emulation if requested.
     */
    if (emulationPrev && PSCbyte1 == 0 && PSCbyte2 == 0 && put_byte == 0x40)
    {
      put_byte |= two_powers[2];
      emulPreventionBits++;
    }
    else
      paddingBits++;

    fwrite(&put_byte, 1, 1, bitstream_bit);

    PSCbyte1 = PSCbyte2;
    PSCbyte2 = put_byte;
    put_byte = 0;
    paddingBits += curr_bit;
    curr_bit = 7;
  }

  return clen;
}

int intraSendMode(int mode)
{
  switch(mode){
  case 0:
    return put_bits(1,0,0,1);
    break;
  case 1:
    return put_bits(2,2,0,1);
    break;
  case 2:
    return put_bits(2,3,0,1);
    break;
  default:
    fprintf(stderr,"Unknown mode in intraSendMode\n");
    exit(1);
    break;
  }
}

int SendPicStart()
{
  return put_bits(22,16,0,0);
}

int SendEndofSeq()
{
  return put_bits(22,31,1,0);
}

/*
 * SendPicHead()
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Function:    
 *    Dumps the picture header information to the bitsream
 * Returns:
 *      Number of bits dumped to the bitstream
 */
int SendPicHead(int IPB,           /* 0- P, 2- I      */
                int FrNo, int sliceMode, int useLoopFilter, int QP,
                int pic_sizex, int pic_sizey,
                int numActiveRefFrames, /* Number of reference frames used  */
                int *orderedRefFrames)  /* Indexes of reference frames in  
                                         * longterm memory                  */
{
  int
    bits=0, FrSz;

  if(pic_sizex==176 && pic_sizey==144) FrSz=0;
  else{
    if(pic_sizex==352 && pic_sizey==288) FrSz=3;
    else{
      if(pic_sizex==128 && pic_sizey==96) FrSz=2;
      else{
        if (pic_sizex % 16 || pic_sizey % 16) {
          fprintf(stderr,"Non multiple of 16 picture size\n");
          fprintf(stdout,"Non multiple of 16 picture size\n");
          exit(-1);
        }
        if (pic_sizex <= 0 || pic_sizex > 2048 ||
            pic_sizey <= 0 || pic_sizey > 2048) {
          fprintf(stderr,"Out of range 2048 picture format in SendPicHead()\n");
          fprintf(stdout,"Out of range 2048 picture format in SendPicHead()\n");
          exit(-2);
        }
        FrSz = 1;
      }
    }
  }

  bits += SendPicStart();            /* Picture Start Code */
  bits += put_bits(1,sliceMode,0,1); /* Slice mode         */
  bits += put_bits(8,FrNo%256,0,1);  /* temporal reference */

  /* It is believed to be reasonable to send picture size
   * for each of the coded frames. This may come useful
   * for graceful quality degradation.
   */

  /* PTYPE start, this field may contain more items in the future */
  if(!IPB)
    bits += put_bits(1,IPB,0,1);   /* P frame */
  else if(IPB == 2)
    bits+=put_bits(2,IPB,0,1);     /* I frame */

  bits+=put_bits(2,FrSz,0,1); /* QCIF, subsampled QCIF, sub-QCIF or CIF */

  if (FrSz == 1)     /* Custom picture size format */
  {
    bits+=put_bits(7,pic_sizex/16-1,0,1);
    bits+=put_bits(7,pic_sizey/16-1,0,1);
  }

  bits+=put_bits(5,QP,0,1); /* frame QP */

  /* Loopfilter information for a P-frame */
  if(IPB != 2)
    bits += put_bits(1, useLoopFilter, 0, 1);
  
  /* In long term memory mode signal reference frames to the decoder */
  if(IPB == 0 && longTermMemSize > 1){

    /* Assume long-term memory with temporal order */

    if(numActiveRefFrames > 1)
      bits += put_bits(1, 1, 0, 1); /* LTM used */
    else
      bits += put_bits(1, 0, 0, 1); /* LTM not used */

  }

  /* Longterm memory size is sent in intra picture header */
  else if(IPB == 2){

    /* Maximum number of reference frames */
    bits += dumpLongtermRVLC(longTermMemSize-1, 1);
  }

  return bits;
}

/*
 * SendSliceHead()
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Function:    
 *    Dumps the Slice Layer header information to the bitsream
 * Returns:
 *      Number of bits dumped to the bitstream
 */

int SendSliceHead(int mbNum, int intraSliceFlag, int sliceQuant,
                  int frNo, int picSizeX, int picSizeY, int dumpSSC,
                  int sendCriticalParams, int IPB /* 0- P, 2- I*/)
{
  int
    sendTR = 1,
    bits = 0, FrSz, mbNumLen;
  /* generate picture size information*/
  if (picSizeX == 176 && picSizeY == 144)
  {
    FrSz = 0;
    mbNumLen = 7;
  }
  else if (picSizeX == 352 && picSizeY == 288)
  {
    FrSz = 3;
    mbNumLen = 9;
  }
  else if (picSizeX == 128 && picSizeY == 96)
  {
    FrSz = 2;
    mbNumLen = 6;
  }
  else
  {
    FrSz = 1;
    mbNumLen = 14;
  }

  if (dumpSSC)
  {
    bits += put_bits(18,3,0,0); /* SSC */

    bits += put_bits(1,sendTR,0,1); /* TR Indicator */

    if (sendTR)
      bits += put_bits(8,frNo%256,0,1);

    /* PTYPE start, this field may contain more items in the future */

    if(!IPB)
      bits += put_bits(1,IPB,0,1);   /* P frame */
    else if(IPB == 2)
      bits += put_bits(2,IPB,0,1);     /* I frame */

    if (IPB == 2)  /* I frame */
    {
      bits += put_bits(1,sendCriticalParams,0,1);

      if (sendCriticalParams)
      {
        /* PFORMAT */

        bits += put_bits(2,FrSz,0,1);

        if (FrSz == 1)
        {
          /* Custom picture size format */

          bits += put_bits(7,picSizeX/16-1,0,1);
          bits += put_bits(7,picSizeY/16-1,0,1);
        }

        bits += put_bits(1,0,0,1);   /* Extended header indicator */

        if ( 0 )
        {
        /* If needed here comes the Extended Header */
        }
      }
    }

    bits += put_bits(5,sliceQuant,0,1); /* Slice QP */
  }

  if (IPB != 2)  /* P  frame only */
    bits += put_bits(1,intraSliceFlag,0,1);

  bits += put_bits(mbNumLen,mbNum,0,1); /* MB address */

  return bits;
}

void SendSliceHeaderIfNeeded(int *currSliceID, int *frameBitsAtStartSlice,
                             int *sliceXmin, int *sliceYmin,
                             int *sliceYmax, int *sliceXmax,
                             int picSizeX, int picSizeY,
                             int mbPerLine, int mbIndex, int qp, int frNo,
                             int sliceBits, Bits *bits)
{

  int sliceCurrBits = get_all_bits() - *frameBitsAtStartSlice;
  
  if ((sliceBits < 0) ? (mbIndex%(mbPerLine*(-sliceBits)) == 0 && mbIndex!=0) :
                        (sliceCurrBits > sliceBits))
  {
    *currSliceID = mbIndex;
    
    /* Stuffing before the next slice SSC */
    put_bits(0,0,1,1);
    
    *frameBitsAtStartSlice = get_all_bits();
    
    /* Dump a new Slice Header (SSC+mbNum+etc) to bitstream */
    
    bits->header += SendSliceHead(*currSliceID, 0, qp, frNo,
                                  picSizeX, picSizeY, 1, 1, 0);
    *sliceXmin    = picSizeX;
    *sliceYmin    = picSizeY;
    *sliceYmax    = 0;
    *sliceXmax    = 0;
    
    fprintf(stdout, "Slice start at P %3d %2d\n", frNo, *currSliceID);
    fprintf(stderr, "Slice start at P %3d %2d\n", frNo, *currSliceID);
  }
}

/*
 * dump8x8MethSel()
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Function:    
 *    Dumps the 8x8 Coding Method information to the bitstream
 * Returns:
 *      Number of bits dumped to the bitstream
 */
static int dump8x8MethSel(int class, int method, int chroma, int mode,
                          int gen_bitstr_flag)
{
  int
    methBits,
    index, ovhNo;

  index    = methSelectNum8x8New(class, method);
  ovhNo    = overheadNumbers8x8[class];
  methBits = SendMethSel_Event(ovhNo, index, gen_bitstr_flag);

  return methBits;
}

/*
 * dump4x4MethSel()
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Function:    
 *    Dumps the 4x4 Coding Method information to the bitstream
 * Returns:
 *      Number of bits dumped to the bitstream
 */
static int dump4x4MethSel(int class, int method, int chroma, int mode,
                          int gen_bitstr_flag)
{
  int
    methBits,
    index, ovhNo;

  index    = methSelectNum4x4New(class, method);
  ovhNo    = overheadNumbers4x4[class];
  methBits = SendMethSel_Event(ovhNo, index, gen_bitstr_flag);

  return methBits;
}


/*
 * dump8x8BlockAcInfo()
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Function:    
 *   Dumps the 8x8 blocks' coefficient layer information to the bitsream
 * Returns:
 *      Number of bits dumped to the bitstream
 */

int dump8x8BlockAcInfo(int mode, int chroma, int qp, int *tmpDouCnt,
                       int *tmpRdcCnt, int *tmpVqCnt, int *tmpTriCnt,
                       int gen_bitstr_flag)
{
  int
    bits = 0, i,
    class8x8, method8x8,
    class4x4, method4x4,
    codedBlks,
    dctAmount, acBits, acBit,
    tableNo, runBits, levelBits,
    last, run, level, index, save4[4];

  class8x8  = region_doublets[*tmpDouCnt].class;
  method8x8 = region_doublets[*tmpDouCnt].transform;
  (*tmpDouCnt)++;
  if(*tmpDouCnt > doublet_counter) {
    printf("tmpDouCnt %d > %d doublet_counter in dump8x8BlockAcInfo()\n",
           *tmpDouCnt, doublet_counter);
    exit(1);
  }
  /* send the 8x8 coding method selected*/
  bits += dump8x8MethSel(class8x8, method8x8, chroma, mode, gen_bitstr_flag);

  if(method8x8 >= METH_4x4_1 && method8x8 <= METH_4x4_15) {
    codedBlks = method8x8 == METH_4x4_15 ? 4 : 1;

    dctAmount = 0;
    for(i=0; i<codedBlks; i++) {
      class4x4  = vq_triplets[*tmpVqCnt].last;
      method4x4 = vq_triplets[*tmpVqCnt].run;
      (*tmpVqCnt)++;
      if(*tmpVqCnt > vq_counter) {
        printf("tmpVqCnt %d > %d vq_counter in dump8x8BlockAcInfo()\n",
               *tmpVqCnt, vq_counter);
        exit(1);
      }
      /* send the 8x8 coding method selected*/
      bits += dump4x4MethSel(class4x4, method4x4, chroma,mode, gen_bitstr_flag);

      if(method4x4 == METH_DCT_4x4 ||
         method4x4 == METH_DCT_4x4_N_0_8 ||
         method4x4 == METH_DCT_4x4_N_4_8 ||
        (method4x4 >= KLT_4x4_N_0_8 &&
         method4x4 <= KLT_4x4_N_7_8)) {
        save4[dctAmount] = method4x4;
        dctAmount++;
      }
    }
  }
  else if((method8x8 >= METH_8x8_DCT && method8x8 <= METH_8x3_DIAG_DCT) ||
          (method8x8 >= METH_KLT_8x8_N_0_8 && method8x8 <= METH_KLT_8x8_TEXT)||
          (method8x8 >= METH_DCT_8x8_N_0_8 && method8x8 <= METH_DCT_8x8_N_4_8)){
    dctAmount = findNumMethods(method8x8, SIZE_8x8);
  }
  else
    dctAmount = 0;

  if(dctAmount > 0) {
    acBits = 0;

    if(method8x8 >= METH_4x4_1 && method8x8 <= METH_4x4_15) {
      index     = 0;
      method4x4 = save4[index++];
    }
    else {
      method4x4 = 0;
    }
    findTripletInfos(method8x8, method4x4, qp, chroma, mode,
                     &tableNo, &runBits, &levelBits);

    while(dctAmount > 0) {
      last  = region_triplets[*tmpTriCnt].last;
      run   = region_triplets[*tmpTriCnt].run;
      level = region_triplets[*tmpTriCnt].level;
      (*tmpTriCnt)++;
      if(*tmpTriCnt > triplet_counter) {
        printf("tmpTriCnt %d > %d triplet_counter in dump8x8BlockAcInfo()\n",
               *tmpTriCnt, triplet_counter);
        exit(1);
      }

      acBit = SendDctOrKlt_EventXxX(last, run, level, tableNo, runBits,
                                    levelBits, gen_bitstr_flag);
      acBits += acBit;

      if(last) {
        dctAmount--;
        if(dctAmount > 0 &&
           method8x8 >= METH_4x4_1 &&
           method8x8 <= METH_4x4_15) {
          method4x4 = save4[index++];
          findTripletInfos(method8x8, method4x4, qp, chroma, mode,
                           &tableNo, &runBits, &levelBits);
        }
      }
    }
    bits += acBits;
  }

  return bits;
}

/*
 * dump4x4BlockAcInfo()
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Function:    
 *   Dumps the 4x4 blocks' coefficient layer information to the bitsream
 * Returns:
 *      Number of bits dumped to the bitstream
 */
int dump4x4BlockAcInfo(int mode, int chroma, int qp, int *tmpRdcCnt,
                       int *tmpVqCnt, int *tmpTriCnt, int gen_bitstr_flag)
{
  int
    bits = 0,
    method8x8,
    class4x4, method4x4,
    codedBlks,
    dctAmount, acBits, acBit,
    tableNo, runBits, levelBits,
    last, run, level;

  method8x8 = METH_4x4_1;
  codedBlks = 1;           /* find4x4Methods(method8x8) */

  dctAmount = 0;
  class4x4  = vq_triplets[*tmpVqCnt].last;
  method4x4 = vq_triplets[*tmpVqCnt].run;
  (*tmpVqCnt)++;
  if(*tmpVqCnt > vq_counter) {
    printf("tmpVqCnt %d > %d vq_counter in dump4x4BlockAcInfo()\n",
           *tmpVqCnt, vq_counter);
    exit(1);
  }

  bits += dump4x4MethSel(class4x4, method4x4, chroma, mode, gen_bitstr_flag);

  if(method4x4 == METH_DCT_4x4 ||
     method4x4 == METH_DCT_4x4_N_0_8 ||
     method4x4 == METH_DCT_4x4_N_4_8 ||
    (method4x4 >= KLT_4x4_N_0_8 &&
     method4x4 <= KLT_4x4_N_7_8)) {
    dctAmount++;
  }

  if(dctAmount > 0) {
    acBits = 0;

    findTripletInfos(method8x8, method4x4, qp, chroma, mode,
                     &tableNo, &runBits, &levelBits);

    while(dctAmount > 0) {
      last  = region_triplets[*tmpTriCnt].last;
      run   = region_triplets[*tmpTriCnt].run;
      level = region_triplets[*tmpTriCnt].level;
      (*tmpTriCnt)++;
      if(*tmpTriCnt > triplet_counter) {
        printf("tmpTriCnt %d > %d triplet_counter in dump4x4BlockAcInfo()\n",
               *tmpTriCnt, triplet_counter);
        exit(1);
      }
              
      acBit = SendDctOrKlt_EventXxX(last, run, level, tableNo, runBits,
                                    levelBits, gen_bitstr_flag);
      acBits += acBit;

      if(last) {
        dctAmount--;
      }
    }

    bits += acBits;
  }

  return bits;
}

/*
 * dumpYInfo()
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Function:    
 *   Dumps the  coefficient layer information for luminance to the bitsream
 * Returns:
 *      Number of bits dumped to the bitstream
 */
int dumpYInfo(int *cnc_y, int xmin, int xmax, int ymin, int ymax, int qp)
{
  int
    bits = 0,
    y, x, tmp = 0,
    tmpTriCnt = 0, tmpDouCnt = 0,
    tmpVqCnt = 0, tmpRdcCnt = 0;

  for (y = ymin; y < ymax; y += 8)
    for (x = xmin; x < xmax; x += 8)
      /* Y coding  */
      if (cnc_y[tmp++])
        bits += dump8x8BlockAcInfo(INTER_BLOCK, LUMA, qp, &tmpDouCnt, &tmpRdcCnt,
                                   &tmpVqCnt, &tmpTriCnt, 1);

  if(tmpTriCnt != triplet_counter_Y ||
     tmpDouCnt != doublet_counter_Y ||
     tmpVqCnt  != vq_counter_Y ||
     tmpRdcCnt != rdc_counter_Y) {
    fprintf(stdout, "tmpTriCnt %d != %d triplet_counter_Y or ",
            tmpTriCnt, triplet_counter_Y);
    fprintf(stdout, "tmpDouCnt %d != %d doublet_counter_Y or ",
            tmpDouCnt, doublet_counter_Y);
    fprintf(stdout, "tmpVqCnt  %d != %d vq_counter_Y or ",
            tmpVqCnt, vq_counter_Y);
    fprintf(stdout, "tmpRdcCnt %d != %d rdc_counter_Y.",
            tmpRdcCnt, rdc_counter_Y);
    nccExit(1);
  }

  return bits;
}

/*
 * dumpUvInfo()
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Function:    
 *   Dumps the  coefficient layer information for chrominance to the bitsream
 * Returns:
 *      Number of bits dumped to the bitstream
 */
int dumpUvInfo(int *cnc_u, int *cnc_v, int regionSize, int qp)
{
  int
    bits = 0,
    tmpTriCnt = triplet_counter_Y, tmpDouCnt = doublet_counter_Y,
    tmpVqCnt = vq_counter_Y, tmpRdcCnt = rdc_counter_Y;

  switch (regionSize) {

    case 16: /* region of interest is a macroblock */
      if (*cnc_u)
        bits += dump8x8BlockAcInfo(INTER_BLOCK, CHROMA, qp, &tmpDouCnt,
                                   &tmpRdcCnt, &tmpVqCnt, &tmpTriCnt, 1);

      if (*cnc_v)
        bits += dump8x8BlockAcInfo(INTER_BLOCK, CHROMA, qp, &tmpDouCnt,
                                   &tmpRdcCnt, &tmpVqCnt, &tmpTriCnt, 1);

      break;

    case 8: /* region of interest is a quadrant */
      if (*cnc_u)
        bits += dump4x4BlockAcInfo(INTER_BLOCK, CHROMA, qp, &tmpRdcCnt,
                                   &tmpVqCnt, &tmpTriCnt, 1);

      if (*cnc_v)
        bits += dump4x4BlockAcInfo(INTER_BLOCK, CHROMA, qp, &tmpRdcCnt,
                                   &tmpVqCnt, &tmpTriCnt, 1);
      break;

    default:
      fprintf(stdout, "Wrong!\n");
      exit(1);
      break;
  }

  if(tmpTriCnt != triplet_counter ||
     tmpDouCnt != doublet_counter ||
     tmpVqCnt  != vq_counter ||
     tmpRdcCnt != rdc_counter) {
    fprintf(stdout, "tmpTriCnt %d != %d triplet_counter or ",
            tmpTriCnt, triplet_counter);
    fprintf(stdout, "tmpDouCnt %d != %d doublet_counter or ",
            tmpDouCnt, doublet_counter);
    fprintf(stdout, "tmpVqCnt  %d != %d vq_counter or ",
            tmpVqCnt, vq_counter);
    fprintf(stdout, "tmpRdcCnt %d != %d rdc_counter.",
            tmpRdcCnt, rdc_counter);
    nccExit(1);
  }

  return bits;
}

/*
 * SendDctOrKlt_EventXxX()
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Function:    
 *   Dumps the  DCT or KLT coefficients to the bitstream.
 * Returns:
 *      Number of bits dumped to the bitstream
 */
int SendDctOrKlt_EventXxX(int last, int run, int level,
                          int tableNo, int runBits, int levelBits,
                          int gen_bitstr_flag)
{
  int
    bits,
    abslevel = abs(level),
    escapeInd, vlcInd;

  vlc_item
    *vlcTable;

  if (!triplet2index_init) initTriplet2Index();

  vlcInd   = escapeInd = (*vectorOftriplet2IndexCubes[tableNo])[0][0][0];
  vlcTable = &((*vectorOfvlcTable[tableNo])[0]);

  if(last >= LAST || run >= SIZE_8x8 || abslevel > LEVEL_10 ||
     last <     0 || run <         0 || abslevel <        1) {
    printf("Wrong last %d, run %d or level %d\n", last, run, level);
    exit(1);
  }

  if (abslevel < LEVEL_05) {
    vlcInd = (*vectorOftriplet2IndexCubes[tableNo])[last][run][abslevel];
  }
    
  bits = SendMccTriplet(last, run, level, vlcTable, vlcInd,
                        levelBits, runBits, escapeInd, gen_bitstr_flag);
  if(bits <= 0 || bits >= 80) {
    printf("Wrong bits %d in SendDctOrKlt_EventXxX()\n", bits);
    printf("last %d run %d level %d\n", last, run, level);
    printf("runBits %d levelBits %d\n", runBits, levelBits);
    printf("tableNo %d vlcInd %d escapeInd %d\n", tableNo, vlcInd, escapeInd);
    fflush(stdout); exit(2);
  }

  return bits;
}


/* Fast version of SendDctOrKlt for use by calcDctBits       */
/* If using this function, it really important to initialize */
/* in the main module first by doing :                       */
/*   initTriplet2Index();                                    */
int SendDctOrKlt_EventXxX_FAST(int last, int run, int level,
                               int tableNo, int runBits,
                               int levelBits, int escapeInd,
                               vlc_item *vlcTable)
{
	register int abslevel = level,
		vlcInd;

	if (abslevel<0)
		abslevel = -abslevel;

	if (abslevel < LEVEL_05) {
		vlcInd = (*vectorOftriplet2IndexCubes
				[tableNo])[last][run][abslevel];

		if (vlcInd != escapeInd)
  			return(vlcTable[vlcInd].len + 1);
		else 
			return(vlcTable[escapeInd].len + 
						1 + runBits + 1 + levelBits);
	}		
	else
	{
		return(vlcTable[escapeInd].len + 
				1 + runBits + 1 + levelBits);
	}
}



/*
 * SendMethSel_Event()
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Function:    
 *   Dumps the  selected coding method information to the bitstream
 * Returns:
 *      Number of bits dumped to the bitstream
 */
int SendMethSel_Event(int tableNo, int rankNo, int gen_bitstr_flag)
{
  int
    bits,
    vlcInd = 14 * tableNo + rankNo;

  if (tableNo < 2 || tableNo > 74)
  {
    if (tableNo == 0)
    {
      if (gen_bitstr_flag)
      {
        printf("Out of range tableNo (%d) in SendMethSel_Event()\n", tableNo);
        exit(2);
      }
      else
      {
        bits = vlcTableMethSel[vlcInd].len;
        return bits;
      }
    }
    else if (tableNo == 1)
    {
      return 0;
    }
    else {
      printf("Out of range tableNo (%d) in SendMethSel_Event()\n", tableNo);
      exit(2);
    }
  }

  if (rankNo < 1 || rankNo > 12)
  {
    printf("Out of range rankNo (%d) in SendMethSel_Event()\n", rankNo);
    exit(2);
  }

  if (gen_bitstr_flag)
  {
    bits = put_vlc_item2(vlcTableMethSel[vlcInd], 0);
  }
  else
  {
    bits = vlcTableMethSel[vlcInd].len;
  }

  return bits;
}


static int SendMccTriplet(int last, int run, int level,
                          vlc_item *vlcTable, int vlcInd,
                          int levelBits, int runBits, int escapeInd,
                          int gen_bitstr_flag)
{
  int
    sign, bits = 0;
    
  if(!level){
    fprintf(stderr,"Zero level of AC in SendMccTriplet()\n");
    exit(1);
  }
      
  sign  = (level > 0) ? 0 : 1;
  level = abs(level);

  if(vlcInd != escapeInd)
  {
    if (gen_bitstr_flag)
    {
      bits += put_vlc_item2(vlcTable[vlcInd], 0);
      bits += put_bits(1, sign, 0, 1);
    }
    else
      bits += vlcTable[vlcInd].len + 1;
  }
  else
  {
    if (gen_bitstr_flag)
    {
      bits += put_vlc_item2(vlcTable[escapeInd], 0); /*escape*/
      bits += put_bits(1, last, 0, 1);
      bits += put_bits(runBits, run, 0, 1);
      bits += put_bits(1, sign, 0, 1);
      bits += put_bits(levelBits, level, 0, 1);
    }
    else
      bits += vlcTable[escapeInd].len + 1 + runBits + 1 + levelBits;
  }

  return bits;
}


static void dcdIndexToRangeIndex(int32 dcdIndex, int32 *rangeDcdIndex)
{
  static unsigned char
    index2RangeIndex[30] = { 0, 1, 2, 3, 4, 4, 5, 5, 6, 6,
                             7, 7, 7, 7, 8, 8, 8, 8, 9, 9,
                             9, 9,10,10,10,10,10,10,10,10
                           };

  if (dcdIndex < 30)
    *rangeDcdIndex = index2RangeIndex[dcdIndex];
  else
    *rangeDcdIndex = 11;
}



void bmIndexToIndex(int bmIndex, int *index)
{
  if     (bmIndex <   2) *index = 0;
  else if(bmIndex <   4) *index = 1;
  else if(bmIndex <   6) *index = 2;
  else if(bmIndex <   8) *index = 3;
  else if(bmIndex <  12) *index = 4;
  else if(bmIndex <  16) *index = 5;
  else if(bmIndex <  24) *index = 6;
  else if(bmIndex <  32) *index = 7;
  else if(bmIndex <  48) *index = 8;
  else if(bmIndex <  64) *index = 9;
  else if(bmIndex <  96) *index = 10;
  else if(bmIndex < 128) *index = 11;
  else if(bmIndex < 192) *index = 12;
  else if(bmIndex < 256) *index = 13;
  else if(bmIndex < 384) *index = 14;
  else                   *index = 15;
}

/*
 * SendSelDcdCncSign_Event()
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Function:    
 *   Dumps the  method coding selection , DC difference and CNC information to the bitstream
 * Returns:
 *      Number of bits dumped to the bitstream
 */
int SendSelDcdCncSign_Event(int ipContextNo, int selRank, int dcdIndex, int cncValue, int sign,
                            int qp, int gen_bitstr_flag)
{
  static unsigned char
    auxLog[5] = {0, 0, 1, 2, 2};

  int32
    bits,
    rangeDcdIndex,
    offsetDcdSubindex,
    bitsDcdSubindex,
	selLenList;


  selLenList = ipContextLists[ipContextNo*(MAX_IP_LIST_SIZE+1)];

  /* Coding the intra prediction method selection information */

  if (selLenList > 6 || selLenList < 1) {
    fprintf(stderr, "Invalid length of list %d\n", (int)selLenList);
    nccExit(1);
  }
  else if (selLenList < 2)
    bits = 0;
  else
  {
    if (selRank == 0)
    {
      if (gen_bitstr_flag)
        bits = put_bits(1, 1, 0, 1);
      else
        bits = 1;
    }
    else if (selRank == 1)
    {
      if (gen_bitstr_flag)
        bits = put_bits(2, 1, 0, 1);
      else
        bits = 2;
    }
    else
    {
      if (gen_bitstr_flag)
        bits = put_bits(2, 0, 0, 1);
      else
        bits = 2;
    }
    if (selLenList > 3 && selRank >= 2)
    {
      if (gen_bitstr_flag)
        bits += put_bits(auxLog[selLenList - 2], selRank - 2, 0, 1);
      else
        bits += auxLog[selLenList - 2];
    }
  }

  /* Saving the CNC bit */
  if (gen_bitstr_flag)
    bits += put_bits(1, ((cncValue) ? 1 : 0) , 0, 1);
  else
    bits += 1;

  /* Saving the DC difference information */
  dcdIndexToRangeIndex(dcdIndex, &rangeDcdIndex);
  dcdRangeIndexToBits(rangeDcdIndex, qp,
                      &offsetDcdSubindex, &bitsDcdSubindex);
  if (gen_bitstr_flag)
  {
    bits += put_vlc_item2(vlcIntraPredDcd[rangeDcdIndex], 0);
    if (bitsDcdSubindex)
      bits += put_bits(bitsDcdSubindex, dcdIndex - offsetDcdSubindex, 0, 1);
  }
  else
  {
    bits += vlcIntraPredDcd[rangeDcdIndex].len;
    if (bitsDcdSubindex)
      bits += bitsDcdSubindex;
  }
  /* If nonzero, send 1 bit for sign */

  if (dcdIndex)
  {
    int
      signCode = (sign > 0) ? 0 : 1;
    if (gen_bitstr_flag)
      bits += put_bits(1, signCode, 0, 1);
    else
      bits += 1;
  }

  return bits;
}


/*
 * SendBMV_Event()
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Function:    
 *   Dumps the  Block Matching vector information to the bitstream
 * Returns:
 *      Number of bits dumped to the bitstream
 */
int SendBMV_Event(int dY, int dX, int reducedSearchRange, int gen_bitstr_flag)
{
  int
    bits,
    BitsXLeft,
    BitsXRight,
    offsetX;

  if (reducedSearchRange)
  {
    BitsXLeft  = 0;
    BitsXRight = 4;
    offsetX    = 8;
  }
  else
  {
    BitsXLeft  =  3;
    BitsXRight =  5;
    offsetX    = 15;
  }

  dY = -dY;
  dX = dX + offsetX;

  /*
   * Send dY first.
   */
  if (gen_bitstr_flag)
    bits = put_bits(4, dY, 0, 1);
  else
    bits = 4;

  /*
   * Send dX second depending on the value of dY.
   */
  if (dY < 8)
  {
    if (BitsXLeft)
      if (gen_bitstr_flag)
        bits += put_bits(BitsXLeft, dX, 0, 1);
      else
        bits += BitsXLeft;
  }
  else
  {
    if (gen_bitstr_flag)
      bits += put_bits(BitsXRight, dX, 0, 1);
    else
      bits += BitsXRight;
  }

  return bits;
}



void ResetBitCount(Bits *bits)
{
  bits->header  =
  bits->mode    =
  bits->motcoef =
  bits->coefy   =
  bits->coefu   =
  bits->coefv   =
  bits->cncy    =
  bits->cncuv   =
  bits->total   = 0;
}
