#ifndef GLOBALS_H
#define GLOBALS_H

#include <stdio.h>
#include <stdlib.h>

#include "nrctypes.h"

#define PCT_INTER    0
#define PCT_INTRA    2
#define PCT_PB       3
#define PCT_END     10
#define PCT_NONE  (-1)

#define LONG_TERM_MEM_MAX_SIZE 10

typedef struct
{
  int header;
  int mode;
  int motcoef;
  int coefy;
  int coefu;
  int coefv;
  int cncy;
  int cncuv;
  int total;
} Bits;

typedef struct
{
  double psnr_y;
  double psnr_u;
  double psnr_v;
} Psnr;


typedef struct {
  double lower;
  double upper;
} conditions;


struct frame
{
  Byte *yptr;
  Byte *uptr;
  Byte *vptr;
};

typedef struct 
{
  Byte regionMode;
  Byte pecMode;
  Byte splitted[LONG_TERM_MEM_MAX_SIZE];

  Byte   tryWithPrediction;
  Byte   zeroRefinement;
  int    predNeighIndex;
  double predMinLagr;

  short no_of_pels;

  short sliceID;

  short xmin;
  short xmax;
  short ymin;
  short ymax;

  Byte predPossible;
  Byte predDir;
  Byte predSel;
  Byte predDirsAvail;
  Byte predSelNeededUp;
  Byte predSelNeededLeft;
  
  Byte zero_motion;

  Byte refFrameNumber;

  double lagr;
  double mseMin[LONG_TERM_MEM_MAX_SIZE];

  double motCoefOrthQuant[6];       /* Final orthonormalized (refinement)
                                     * parameters. These are the ones actually
                                     * send to channel/decoded              */
  double motCoefEstim[LONG_TERM_MEM_MAX_SIZE][6];
                                    /* Estimated motion parameters          */
  double motCoefEstimLoc[LONG_TERM_MEM_MAX_SIZE][6];
                                    /* Estimated motion parameters          */
  double motCoefPredRefinedGlob[6];
  double motCoefPredGlob[6];

  int
    motCoefOrth[6];   /* Final orthonormalized (refinement)
                       * parameters. These are the ones actually
                       * send to channel/decoded
                       */
  
  int32
    affineData[6];    /* This array contains following items in fixed point
                       * (16 bits after the point):
                       *
                       * affineData[0] Horizontal component of the motion
                       *                vector in the upper left corner of
                       *                the (macro)block.
                       * affineData[1] Delta vector to be added to the hor mv
                       *                when stepping one pixel down
                       * affineData[2] Delta vector to be added to the hor mv
                       *                when stepping one pixel to the right
                       *
                       * affineData[3] Same data for the vertical component
                       * affineData[4]  of the motion vector.
                       * affineData[5]
                       */
  
  int32
    affineDataPred[6];/* Predicted affine parameters in the same format
                       * as affineData[6]
                       */
  
  /* partial luminance INTRA lagrangian data for 4 8x8 subblocks of an 16x16 block */
  int bitsIntra[4]; 
  int errIntra[4]; 
  char cncIntra[4];

  /* partial luminance COPY lagrangians for 4 8x8 subblocks of an 16x16 block */
  int bitsCopy[4]; 
  int errCopy[4]; 
  char cncCopy[4];

  /* Saved BM vectors for subblocks */
  int BMvectors[2*4];
} object;

typedef struct 
{
  Byte  regionMode;
  short sliceID;
  int   xMin, yMin;
  int32 affineData[6];
  Byte  pecMode; 
  Byte  refFrame;
} objectBuffer_t;



void ResetBitCount(Bits *bits);

#endif
