#include <string.h>
#include <assert.h>

#include "mainFlags.h"
#include "IOimg.h"
#include "mcp.h"
#include "motionOpt.h"
#include "motionPred.h"
#include "PredErrorCoder.h"
#include "encMux.h"

extern int longTermMemSize;

typedef struct 
{
  Byte regionMode;

  Byte predPossible;
  Byte predDir;
  Byte predSel;
  Byte predDirsAvail;
  Byte predSelNeededUp;
  Byte predSelNeededLeft;

  Byte refFrameNumber;

  int
    motCoefOrth[6];   /* Final orthonormalized (refinement)
                       * parameters. These are the ones actually
                       * send to channel/decoded
                       */
  
  int32
    affineDataPred[6];/* Predicted affine parameters in the same format
                       * as affineData[6]
                       */
  
} ltmObject;

double mseRegionLum(double A[], object *region, Byte *curr, Byte *prev);
void InitialMotionRefinement(object *currRegion,
                             object *objectList,
                             int regionIndex,
                             struct frame *currFr,
                             struct frame *prevFr,
                             int picSizeX, int picSizeY, 
                             int aQp, double lambda,
                             int bitsOverheadUp,
                             int bitsOverheadLeft,
                             int complexityMode,
                             int refFrame);
void ioiAllocPic(u_char **picptr, int32 picSizeX, int32 picSizeY);

/***************************************************************************
 * Function    : Optimize_Motion
 * Purpose     : Estimate motion coefficients of a block. Estimation is
 *               done using a sequence of Gauss-Newton iterations.
 ***************************************************************************/

void motOptimizeMotion(object *currRegion,
                       struct frame *currFr,
                       struct frame *prevFr,
                       Byte *subsCurrY,
                       Byte *subsPrevY,
                       int  picSizeX, int  picSizeY, 
                       int stage, int qp, double lambda,
                       int complexityMode, int *BMvectors,
                       int refFrame)
{

  Byte
    *currY = currFr->yptr,
    *currU = currFr->uptr,
    *currV = currFr->vptr,
    *prevY = prevFr->yptr,
    *prevU = prevFr->uptr,
    *prevV = prevFr->vptr,
    zeroMotion;

  int 
    j;

  double
    motCoefGlob[6];

  /* Invoke motion estimation function */
  if(!currRegion->zero_motion){
  
    Initial_Motion_Estimation(currRegion, currY, prevY, currU, prevU,
                              currV, prevV, subsCurrY, subsPrevY,
                              picSizeX, picSizeY, stage, qp, lambda,
                              &zeroMotion,
                              currRegion->motCoefEstim[refFrame], /* Local! */
                              &currRegion->mseMin[refFrame], complexityMode,BMvectors);

    if(zeroMotion)
      currRegion->zero_motion = 1;
  }
  else{
    for(j = 0; j < 6; j++)
      currRegion->motCoefEstim[refFrame][j] = 0.0;
    currRegion->mseMin[refFrame] = currRegion->mseMin[refFrame-1] + 100.0;
  }

  memcpy(currRegion->motCoefEstimLoc[refFrame],
         currRegion->motCoefEstim[refFrame], 6*sizeof(double));

  /* Transform coeffients to global system */
  mcpTransformLoc2GlobCoeffs(currRegion->motCoefEstim[refFrame], currRegion,
                             picSizeX, picSizeY, motCoefGlob);
  
  for(j = 0; j < 6; j++)
    currRegion->motCoefEstim[refFrame][j] = motCoefGlob[j];
  
  currRegion->refFrameNumber = 0;
  
}

void motPrepareSubsampledImages(Byte **prevSubsampled,
                                Byte **currSubsampled,
                                struct frame **longTermMem,
                                struct frame *currFr,
                                int picSizeX, int picSizeY)
{

  static int firstInvo = 1;
  int        i, xInd, yInd, subsInd = 0;
  Byte       *tmpBytePtr, *currY, *prevY;

  if(firstInvo){

    firstInvo = 0;

    /* Allocate memory for subsampled reference frames and subsampled
     * current frame
     */
    for(i = 0;i < longTermMemSize;i++)
       ioiAllocPic(&prevSubsampled[i], picSizeX/2, picSizeY/2);

    ioiAllocPic(currSubsampled, picSizeX/2, picSizeY/2);

  }

  /* Rotate the buffer where subsampled reference frames are stored */
  tmpBytePtr = prevSubsampled[longTermMemSize-1];
  for(i = longTermMemSize-1;i > 0;i--)
    prevSubsampled[i] = prevSubsampled[i-1];
  prevSubsampled[0] = tmpBytePtr;

  currY = currFr->yptr;
  prevY = longTermMem[0]->yptr;

  for(yInd = 0;yInd < picSizeY;yInd += 2){

    for(xInd = 0;xInd < picSizeX;xInd += 2){
      (*currSubsampled)[subsInd]   = currY[xInd];
      prevSubsampled[0][subsInd++] = prevY[xInd];
    }

    currY += 2*picSizeX;
    prevY += 2*picSizeX;
  }

}

void motInitialEstimation32x32(object *objectList,
                               struct frame *currFr,
                               struct frame **longTermMem,
                               Byte *currSubsampled,
                               Byte **prevSubsampled,
                               int posX, int posY,
                               int picSizeX, int picSizeY,
                               int numRefFrames, int qp,
                               double lambda,
                               int complexityMode, int *BMvectors)
{

  object tmpRegion32x32;
    object* outRegion;
  
  int mbPerLine,mbIndex,refFrame;

  /* Create 32x32 region */
  tmpRegion32x32.zero_motion= 0;
  tmpRegion32x32.xmin       = posX;
  tmpRegion32x32.ymin       = posY;
  tmpRegion32x32.xmax       = ((posX+32 > picSizeX) ? picSizeX-1 : posX+31);
  tmpRegion32x32.ymax       = ((posY+32 > picSizeY) ? picSizeY-1 : posY+31);
  tmpRegion32x32.no_of_pels = ((tmpRegion32x32.xmax -
                                tmpRegion32x32.xmin + 1) *
                               (tmpRegion32x32.ymax -
                                tmpRegion32x32.ymin + 1));


  /* Motion estimation (for all the reference frames) */
  for(refFrame = 0;refFrame < numRefFrames;refFrame++){
    motOptimizeMotion(&tmpRegion32x32, currFr, longTermMem[refFrame],
                      currSubsampled, prevSubsampled[refFrame], picSizeX, picSizeY,
                      1, qp, lambda, complexityMode, BMvectors, refFrame);

    /* Creating macroblocks */
    split_one_label(&tmpRegion32x32, objectList, 16, 16,
                    picSizeX, picSizeY, refFrame);

  }

  /* Put the MB vectors calculated in 32x32 stage into 16x16 object list */
  
  mbPerLine = picSizeX/16;
  mbIndex   = tmpRegion32x32.xmin/16 + (tmpRegion32x32.ymin/16)*mbPerLine;

  /*** REGION 1 - UPPER LEFT REGION ***/

  
  outRegion = &objectList[5*mbIndex];

  outRegion->BMvectors[2*(0)]         = BMvectors[2*(4*0+0)];
  outRegion->BMvectors[2*(0)+1]       = BMvectors[2*(4*0+0)+1];
  outRegion->BMvectors[2*(1)]         = BMvectors[2*(4*0+1)];
  outRegion->BMvectors[2*(1)+1]       = BMvectors[2*(4*0+1)+1];
  outRegion->BMvectors[2*(2)]         = BMvectors[2*(4*1+0)];
  outRegion->BMvectors[2*(2)+1]       = BMvectors[2*(4*1+0)+1];
  outRegion->BMvectors[2*(3)]         = BMvectors[2*(4*1+1)];
  outRegion->BMvectors[2*(3)+1]       = BMvectors[2*(4*1+1)+1];


  /*** REGION 2 - UPPER RIGHT REGION ***/

  if ((tmpRegion32x32.xmin + 2*16 - 1) < picSizeX){
      
  outRegion = &objectList[5*(mbIndex+1)];

  outRegion->BMvectors[2*(0)]         = BMvectors[2*(4*0+2)];
  outRegion->BMvectors[2*(0)+1]       = BMvectors[2*(4*0+2)+1];
  outRegion->BMvectors[2*(1)]         = BMvectors[2*(4*0+3)];
  outRegion->BMvectors[2*(1)+1]       = BMvectors[2*(4*0+3)+1];
  outRegion->BMvectors[2*(2)]         = BMvectors[2*(4*1+2)];
  outRegion->BMvectors[2*(2)+1]       = BMvectors[2*(4*1+2)+1];
  outRegion->BMvectors[2*(3)]         = BMvectors[2*(4*1+3)];
  outRegion->BMvectors[2*(3)+1]       = BMvectors[2*(4*1+3)+1];

  }

  /*** REGION 3 - LOWER LEFT REGION ***/

  if ((tmpRegion32x32.ymin + 2*16-1) < picSizeY){

  outRegion = &objectList[5*(mbIndex+mbPerLine)];
  
  outRegion->BMvectors[2*(0)]         = BMvectors[2*(4*2+0)];
  outRegion->BMvectors[2*(0)+1]       = BMvectors[2*(4*2+0)+1];
  outRegion->BMvectors[2*(1)]         = BMvectors[2*(4*2+1)];
  outRegion->BMvectors[2*(1)+1]       = BMvectors[2*(4*2+1)+1];
  outRegion->BMvectors[2*(2)]         = BMvectors[2*(4*3+0)];
  outRegion->BMvectors[2*(2)+1]       = BMvectors[2*(4*3+0)+1];
  outRegion->BMvectors[2*(3)]         = BMvectors[2*(4*3+1)];
  outRegion->BMvectors[2*(3)+1]       = BMvectors[2*(4*3+1)+1];

  }

  /*** REGION 4 - LOWER RIGHT REGION ***/

  if ((tmpRegion32x32.xmin+2*16-1) < picSizeX && 
      (tmpRegion32x32.ymin+2*16-1) < picSizeY){

  outRegion = &objectList[5*(mbIndex+mbPerLine+1)];
  
  outRegion->BMvectors[2*(0)]         = BMvectors[2*(4*2+2)];
  outRegion->BMvectors[2*(0)+1]       = BMvectors[2*(4*2+2)+1];
  outRegion->BMvectors[2*(1)]         = BMvectors[2*(4*2+3)];
  outRegion->BMvectors[2*(1)+1]       = BMvectors[2*(4*2+3)+1];
  outRegion->BMvectors[2*(2)]         = BMvectors[2*(4*3+2)];
  outRegion->BMvectors[2*(2)+1]       = BMvectors[2*(4*3+2)+1];
  outRegion->BMvectors[2*(3)]         = BMvectors[2*(4*3+3)];
  outRegion->BMvectors[2*(3)+1]       = BMvectors[2*(4*3+3)+1];
  }
	


}

void motEstimateMacroblock(object *objectList,
                           struct frame *currFr,
                           struct frame *prevFr,
                           Byte *subsCurrY,
                           Byte *subsPrevY,
                           int mbPosX, int mbPosY,
                           int picSizeX, int picSizeY,
                           int qp, double lambda,
                           int numRefFrames,
                           int complexityMode, int *BMvectors,
                           int refFrame)
{

  int
    mbPerLine = picSizeX/16,
    mbIndex   = mbPosX/16 + (mbPosY/16)*mbPerLine,
    blkIndex;

  double
    mse,
    mseMax,
    splitThr;

  object
    *currRegion16, *currRegion8;

  if(complexityMode == COMPLEXITY_MODE_ANCHOR ||
     complexityMode == COMPLEXITY_MODE_PRACTICAL){
    splitThr = lambda*20/256 + qp*qp/3;
  }
  else{
    splitThr = 2.0*(lambda*20/256 + qp*qp/3);
    if(splitThr < 40.0)
      splitThr = 40.0;
  }

  /*
   * Motion estimation for 16x16 macroblocks
   */

  currRegion16 = &objectList[5*mbIndex];

  motOptimizeMotion(currRegion16, currFr, prevFr,
                    subsCurrY, subsPrevY, picSizeX, picSizeY,
                    2, qp, lambda, complexityMode, BMvectors,
                    refFrame);

  /* If 16x16 motion compensation was not able to handle the
   * regions well enough split it into four and perform
   * motion estimation (i.e. if approximated lagrangian is
   * larger than the approximated splitting cost,
   * As quantizationErrorEstimate LMSE estimate for quantization
   * squared error is utilized. This is a bit pessimistic but a
   * good trade-off between complexity and performance.)
   */

  if(currRegion16->mseMin[refFrame] > splitThr/4 &&
     currRegion16->mseMin[refFrame] < 1.5*currRegion16->mseMin[0]){

    split_one_label(currRegion16, objectList, 8, 8,
                    picSizeX, picSizeY, refFrame);

    /* Check all the splitted blocks with inherited parameters,
     * to see if mse(8x8Blk) > splitThr
     */
    mseMax = 0.0;

    if(currRegion16->mseMin[refFrame] >  splitThr/4 &&
       currRegion16->mseMin[refFrame] <= splitThr){

      for(blkIndex = 1;blkIndex < 5;blkIndex++){

        currRegion8 = currRegion16 + blkIndex;

        mse = mseRegionLum(currRegion8->motCoefEstim[refFrame],
                           currRegion8, currFr->yptr, prevFr->yptr);

        if(mse > mseMax)
          mseMax = mse;

        if(mseMax > splitThr)
          break;
      }
    }

    if(currRegion16->mseMin[refFrame] > splitThr ||
       mseMax                         > splitThr){

      /*
       * Motion estimation for 8x8 blocks
       */

      for(blkIndex = 1;blkIndex < 5;blkIndex++){

        currRegion8 = currRegion16 + blkIndex;

        currRegion8->BMvectors[0] = currRegion16->BMvectors[2*(blkIndex-1)];
        currRegion8->BMvectors[1] = currRegion16->BMvectors[2*(blkIndex-1)+1];

        motOptimizeMotion(currRegion8, currFr, prevFr,
                          subsCurrY, subsPrevY, picSizeX, picSizeY,
                          3, qp, lambda, complexityMode, BMvectors,
                          refFrame);

      } 
	  
    }
    else{
      currRegion16->splitted[refFrame] = 0;
    }
  }
}

void motFindMode(object *currRegion, object *objectList, int regionIndex,
                 double *motCoefGlob, int *motCoefOrthQuant,
                 struct frame *currFr, struct frame *prevFr,
                 struct frame *predBuffer, int picSizeX, int picSizeY,
                 int qp, int aQP, double lambda,  int complexityMode,
                 int refFrame)
{

  int
    bitsDir,
    bitsSelectionUpper,
    bitsSelectionLeft,
    regionSize = currRegion->no_of_pels == 256 ? 16 : 8,
    mbPerLine  = picSizeX/16,
    motionNeighs,
    checkMotionParameterPrediction;

  /* Find prediction alternatives */
  mcpFindPredAlternatives(objectList, regionIndex, regionSize,
                          currRegion->xmin,
                          currRegion->ymin, mbPerLine,
                          &currRegion->predPossible,
                          &currRegion->predDirsAvail,
                          &currRegion->predDir,
                          &currRegion->predSelNeededUp,
                          &currRegion->predSelNeededLeft,
                          refFrame, &motionNeighs);

  bitsDir            = currRegion->predDirsAvail     == PREDDIR_BOTH   ? 1:0;
  bitsSelectionUpper = currRegion->predSelNeededUp   == PREDSEL_NEEDED ? 1:0;
  bitsSelectionLeft  = currRegion->predSelNeededLeft == PREDSEL_NEEDED ? 1:0;

  currRegion->zeroRefinement = 1;

  /*
   * This is a special case with multiple reference frames. Motion parameter
   * prediction can take place only from neighbors with same reference frame.
   * Because reference frame infomation for the macroblock is transmitted
   * after the mode information (which is dependent on the information whether
   * prediction is available or not) currRegion->predPossible is forced to
   * 'true' if prediction is possible with any of the reference frames.
   */

  checkMotionParameterPrediction = currRegion->predPossible;

  if(regionSize == 16 && motionNeighs)
    currRegion->predPossible = 1;

  if(checkMotionParameterPrediction)
    InitialMotionRefinement(currRegion, objectList, regionIndex,
                            currFr, prevFr,
                            picSizeX, picSizeY, aQP, lambda,
                            bitsDir+bitsSelectionUpper,
                            bitsDir+bitsSelectionLeft, complexityMode, refFrame);

  /*
   * Code motion without prediction
   */

  currRegion->lagr =
    coeffRem(currRegion,objectList, regionIndex, motCoefGlob, NULL, motCoefOrthQuant, NULL,
             currFr, prevFr, picSizeX, picSizeY, qp, aQP, lambda,
             &currRegion->regionMode, &currRegion->pecMode, 0, 0);
  
  if(regionSize == 16) setLagrangianCopyRegionFlag(1);
  
  /*
   * Store the coefficients offering the best lagrangian to
   * currRegion16->affineData. This is used when predicting
   * other macroblocks
   */
  mcpOrthToMvDelta(motCoefOrthQuant, currRegion->affineData,
                   regionSize);

  /*
   * Lagrangian for intra mode
   */

  if(refFrame == 0 || regionSize == 8){

    if(!currRegion->zero_motion || !currRegion->zeroRefinement){

      int    bits;
      u_char pecMode;
      double rate =
        lagrSegmentCost(prevFr, currFr, currRegion, objectList,regionIndex,NULL, NULL,
                        picSizeX, picSizeY, lambda, qp, aQP,
                        REGMODE_INTRA, PREDDIR_NOTHING,
                        1, &pecMode, &bits);

	    if(regionSize==16) setLagrangianFlag(1);

      if (rate < currRegion->lagr){
        currRegion->lagr       = rate;
        currRegion->regionMode = REGMODE_INTRA;
        currRegion->pecMode    = pecMode;
      }
    }
  }

  /*
   * Prediction from Up and Left.
   */
  if(checkMotionParameterPrediction)
    motionPred(currRegion, objectList, regionIndex, currFr,
               prevFr, predBuffer, qp, aQP, lambda, picSizeX, picSizeY,
               complexityMode);

}

void motFindFinalMotionAndMode(object *objectList,
                               struct frame *currFr,
                               struct frame **longTermMem,
                               struct frame *predFr,
                               struct frame *predBuffer,
                               int mbPosX, int mbPosY,
                               int picSizeX, int picSizeY,
                               int qp, int aQP, double lambda,
                               int *lagrNoLTM,
                               int currSliceID,
                               int complexityMode,
                               int numRefFrames)
{

  int
    mbPerLine  = picSizeX/16,
    mbIndex    = mbPosX/16 + (mbPosY/16)*mbPerLine,
    blkIndex,
    regionIndex,
    indSplitted,
    refFrameBest = 0,
    refFrame     = 0;

  object
    *currRegion16, *currRegion8;

  double
    lagrSplitted, lagrBest = HUGE;

  ltmObject
    ltmRegions[5];

  regionIndex           = 5*mbIndex;
  currRegion16          = &objectList[regionIndex];
  currRegion16->sliceID = currSliceID;

  /*
   * 16x16 region
   */

  setLagrangianFlag(0);

  for(refFrame = 0;refFrame < numRefFrames;refFrame++){

    setLagrangianCopyRegionFlag(0);

    motFindMode(currRegion16, objectList, regionIndex,
                currRegion16->motCoefEstim[refFrame],
                currRegion16->motCoefOrth,
                currFr, longTermMem[refFrame], predBuffer,
                picSizeX, picSizeY, qp, aQP, lambda, 
                complexityMode, refFrame);

    currRegion16->lagr += lambda*(dumpLongtermRVLC(refFrame, 0) - 1);

    if(refFrame == 0)
      *lagrNoLTM = (int)(currRegion16->lagr - lambda);

    if(currRegion16->lagr < lagrBest){

      refFrameBest   = refFrame;
      lagrBest       = currRegion16->lagr;

      /* Store the data */
      ltmRegions[0].predPossible      = currRegion16->predPossible;
      ltmRegions[0].predDirsAvail     = currRegion16->predDirsAvail;
      ltmRegions[0].predDir           = currRegion16->predDir;
      ltmRegions[0].predSel           = currRegion16->predSel;
      ltmRegions[0].predSelNeededUp   = currRegion16->predSelNeededUp;
      ltmRegions[0].predSelNeededLeft = currRegion16->predSelNeededLeft;
      
      ltmRegions[0].regionMode        = currRegion16->regionMode;
      
      memcpy(ltmRegions[0].motCoefOrth,    currRegion16->motCoefOrth,    6 * sizeof(int));
      memcpy(ltmRegions[0].affineDataPred, currRegion16->affineDataPred, 6 * sizeof(int32));

    }
    
    /*
     *  8x8 blocks
     */

    if(currRegion16->splitted[refFrame]){
      
      lagrSplitted = (lambda)*SendRegionMode(REGMODE_SPLITTED,
                                             PECMODE_SKIPPED,
                                             currRegion16->predPossible,
                                             16, 0);

      for (blkIndex = 1;blkIndex < 5;blkIndex++){
        
        indSplitted = 5*mbIndex + blkIndex;
        currRegion8 = &objectList[indSplitted];

        currRegion8->refFrameNumber = refFrame;

        /* Mode for this 8x8 block */
        
        motFindMode(currRegion8, objectList, indSplitted,
                    currRegion8->motCoefEstim[refFrame],
                    currRegion8->motCoefOrth,
                    currFr, longTermMem[refFrame], predBuffer,
                    picSizeX, picSizeY, qp, aQP, lambda,
                    complexityMode, refFrame);
        
        lagrSplitted += currRegion8->lagr;
      }

      lagrSplitted += lambda*(dumpLongtermRVLC(refFrame, 0) - 1);

      if (lagrSplitted < lagrBest){

        if(refFrame == 0)
          *lagrNoLTM = (int)(lagrSplitted - lambda);

        ltmRegions[0].regionMode = REGMODE_SPLITTED;
        
        lagrBest       = lagrSplitted;
        refFrameBest   = refFrame;
        
        for (blkIndex = 1;blkIndex < 5;blkIndex++){
          
          indSplitted = 5*mbIndex + blkIndex;
          currRegion8 = &objectList[indSplitted];
          
          /* Store the data */
          ltmRegions[blkIndex].predPossible      = currRegion8->predPossible;
          ltmRegions[blkIndex].predDirsAvail     = currRegion8->predDirsAvail;
          ltmRegions[blkIndex].predDir           = currRegion8->predDir;
          ltmRegions[blkIndex].predSel           = currRegion8->predSel;
          ltmRegions[blkIndex].predSelNeededUp   = currRegion8->predSelNeededUp;
          ltmRegions[blkIndex].predSelNeededLeft = currRegion8->predSelNeededLeft;
          
          ltmRegions[blkIndex].regionMode        = currRegion8->regionMode;
          
          memcpy(ltmRegions[blkIndex].motCoefOrth,    currRegion8->motCoefOrth,    6 * sizeof(int));
          memcpy(ltmRegions[blkIndex].affineDataPred, currRegion8->affineDataPred, 6 * sizeof(int32));
        }
      }
    }
    else{
      lagrSplitted = currRegion16->lagr + 1;
    }
  }

  /* Restore the motion information for the reference frame chosen */
  if(ltmRegions[0].regionMode != REGMODE_SPLITTED){

    currRegion16->refFrameNumber    = refFrameBest;

    currRegion16->predPossible      = ltmRegions[0].predPossible;
    currRegion16->predDirsAvail     = ltmRegions[0].predDirsAvail;
    currRegion16->predDir           = ltmRegions[0].predDir;
    currRegion16->predSel           = ltmRegions[0].predSel;
    currRegion16->predSelNeededUp   = ltmRegions[0].predSelNeededUp;
    currRegion16->predSelNeededLeft = ltmRegions[0].predSelNeededLeft;
    
    currRegion16->regionMode        = ltmRegions[0].regionMode;
    
    memcpy(currRegion16->motCoefOrth,    ltmRegions[0].motCoefOrth,    6 * sizeof(int));
    memcpy(currRegion16->affineDataPred, ltmRegions[0].affineDataPred, 6 * sizeof(int32));

    mcpCalcFinalParams(currRegion16);

  }
  else{

    currRegion16->refFrameNumber = refFrameBest;
    currRegion16->regionMode     = REGMODE_SPLITTED;

    for (blkIndex = 1;blkIndex < 5;blkIndex++){
        
      indSplitted = 5*mbIndex + blkIndex;
      currRegion8 = &objectList[indSplitted];

      currRegion8->refFrameNumber    = refFrameBest;

      currRegion8->predPossible      = ltmRegions[blkIndex].predPossible;
      currRegion8->predDirsAvail     = ltmRegions[blkIndex].predDirsAvail;
      currRegion8->predDir           = ltmRegions[blkIndex].predDir;
      currRegion8->predSel           = ltmRegions[blkIndex].predSel;
      currRegion8->predSelNeededUp   = ltmRegions[blkIndex].predSelNeededUp;
      currRegion8->predSelNeededLeft = ltmRegions[blkIndex].predSelNeededLeft;
      
      currRegion8->regionMode        = ltmRegions[blkIndex].regionMode;
      
      memcpy(currRegion8->motCoefOrth,    ltmRegions[blkIndex].motCoefOrth,    6 * sizeof(int));
      memcpy(currRegion8->affineDataPred, ltmRegions[blkIndex].affineDataPred, 6 * sizeof(int32));

      mcpCalcFinalParams(currRegion8);

    }
  }

  currRegion16->lagr = lagrBest;

}


void motBuildMBPrediction(object *currRegion16,
                          struct frame *prevFr,
                          struct frame *predFr,
                          struct frame *predBuffer,
                          int picSizeX, int picSizeY)
{
  int
    blkIndex,
    ii, jj,
    coord, coordUV,
    yLen, xLen;

  object
    *currRegion8;

  Byte
    *yPtrPred, *yPtrPredBuf,
    *uPtrPred, *uPtrPredBuf,
    *vPtrPred, *vPtrPredBuf;

  /* Build the prediction for this macroblock */
  if(currRegion16->regionMode == REGMODE_SPLITTED){

    for(blkIndex = 1;blkIndex < 5;blkIndex++) {

      currRegion8 = currRegion16 + blkIndex;
      coord       = (((blkIndex - 1) & 1) << 3) | (((blkIndex - 1) & 2) << 6);
      coordUV     = (((blkIndex - 1) & 1) << 2) | (((blkIndex - 1) & 2) << 4);

      if(currRegion8->regionMode != REGMODE_INTRA){

        mcpBuildSegmPred(picSizeX, picSizeY, prevFr->yptr, 
                         prevFr->uptr, prevFr->vptr, currRegion8,
                         &predBuffer->yptr[coord], &predBuffer->uptr[coordUV], 
                         &predBuffer->vptr[coordUV]);
      }
      else{

       for(ii = 0; ii < 8; ii++){
          for(jj = 0; jj < 8; jj++){
            predBuffer->yptr[coord + (ii << 4) + jj] = 0;
            if(!(ii & 1) && !(jj & 1)) {
              predBuffer->uptr[coordUV + (ii << 2) + (jj >> 1)] = 0;
              predBuffer->vptr[coordUV + (ii << 2) + (jj >> 1)] = 0;
            }
          }
        }         
      }
    }
  }
  else if(currRegion16->regionMode != REGMODE_INTRA){

    mcpBuildSegmPred(picSizeX, picSizeY, prevFr->yptr, 
                     prevFr->uptr, prevFr->vptr, currRegion16,
                     predBuffer->yptr, predBuffer->uptr, predBuffer->vptr);
  }
  else{

    for(ii = 0; ii < 16; ii++){
      for(jj = 0; jj < 16; jj++){
        predBuffer->yptr[(ii << 4) + jj] = 0;
        if(!(ii & 1) && !(jj & 1)){
          predBuffer->uptr[(ii << 2) + (jj >> 1)] = 0;
          predBuffer->vptr[(ii << 2) + (jj >> 1)] = 0;
        }
      }
    }  
  }

  /* Copy prediction to the prediction frame */

  /* Width of region is assumed to be multiple of 4 */
  assert(((currRegion16->xmax - currRegion16->xmin + 1) & 3) == 0 &&
         sizeof(Byte) == 1);

  yPtrPredBuf = predBuffer->yptr;
  yPtrPred = predFr->yptr + currRegion16->ymin * picSizeX +
             currRegion16->xmin;
  yLen = currRegion16->ymax - currRegion16->ymin;
  xLen = currRegion16->xmax - currRegion16->xmin;

  for(ii = 0; ii <= yLen; ii++, yPtrPredBuf += 16, yPtrPred += picSizeX){
    for(jj = 0; jj <= xLen; jj += 4){
      yPtrPred[jj] = yPtrPredBuf[jj];
      yPtrPred[jj+1] = yPtrPredBuf[jj+1];
      yPtrPred[jj+2] = yPtrPredBuf[jj+2];
      yPtrPred[jj+3] = yPtrPredBuf[jj+3];
    }
  }

  uPtrPredBuf = predBuffer->uptr;
  uPtrPred = predFr->uptr + ((currRegion16->ymin * picSizeX) >> 2) +
             (currRegion16->xmin >> 1);
  vPtrPredBuf = predBuffer->vptr;
  vPtrPred = predFr->vptr + ((currRegion16->ymin * picSizeX) >> 2) +
             (currRegion16->xmin >> 1);
  yLen >>= 1; xLen >>= 1;

  for(ii = 0; ii <= yLen; ii++, uPtrPredBuf += 8, uPtrPred += picSizeX >> 1,
    vPtrPredBuf += 8, vPtrPred += picSizeX >> 1){
    for(jj = 0; jj <= xLen; jj += 2){
      uPtrPred[jj] = uPtrPredBuf[jj];
      uPtrPred[jj+1] = uPtrPredBuf[jj+1];
      vPtrPred[jj] = vPtrPredBuf[jj];
      vPtrPred[jj+1] = vPtrPredBuf[jj+1];
    }
  }

}
