#ifndef _MVC_H_
#define _MVC_H_

/*
 * mvc.h
 *
 *      Copyright (c) 1997 Nokia Research Center, Tampere, FINLAND
 *
 * Project:
 *      Nokia Very Low Bitrate Codec
 *
 * Contains:
 *      Functions for motion compensated prediction
 */

#include "nrctypes.h"

/*
 * Defines
 */

#define MVC_BOUNDARY_MARGIN   2.001
#define MVC_INTERV            1000
#define MVC_SIZE_SPLINE       4000
#define MVC_STEP_SPLINE       0.001

#define MVC_BOUNDARY_MARGIN2   2.004
#define MVC_INTERV2            256
#define MVC_SIZE_SPLINE2       1024
#define MVC_STEP_SPLINE2       0.004
 
/*
 * Structs and typedefs
 */

/*
 * Function prototypes
 */

/*
 * mvcBuildPredRegionYUV:
 *
 * Parameters:
 *      currRegion  I: Attribute list for segment
 *      prevY       I: Y component of previous reconstructed frame
 *      predY       O: Y component of prediction frame
 *      prevU       I: U component of previous reconstructed frame
 *      predU       O: U component of prediction frame
 *      prevV       I: V component of previous reconstructed frame
 *      predV       O: V component of prediction frame
 *      picSizeX    I: Frame width as pixels
 *      picSizeY    I: Frame height as pixels
 *
 * Function:
 *      For an INTER-type segment, builds the part of prediction frame
 *      that is within segment boundaries, using motion compensation.
 *      The motion coefficients associated with the segment are assumed
 *      to be for global coordinate system.
 *
 * Returns:
 *      Nothing
 */

void mvcBuildPredRegionYUV(object *currRegion, int32 affineData[6],
                           u_char *prevY, u_char *predY,
                           u_char *prevU, u_char *predU, 
                           u_char *prevV, u_char *predV,
                           int32 picSizeX, int32 picSizeY);


/*
 *
 * mvcInitKernel:
 *
 * Parameters:
 *      Nothing
 *
 * Function:
 *      Initializes kernel for the cubic spline interpolation.
 *
 * Returns:
 *      Nothing
 */

void mvcInitKernel(void);

#endif

