#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "mainFlags.h"
#include "tools.h"
#include "mcp.h"
#include "mvc.h"
#include "encMux.h"
#include "assert.h"


#define LAGR_USE_DCT 1

static void lagrPredRegion(object *region, int *motCoefQuant,
                           int32 *affineDataPred,
                           struct frame *prevfr, struct frame *predfr,
                           int picSizeX, int picSizeY,
                           int regionMode)
{
  int
    j, regionSize = region->no_of_pels == 64 ? 8 : 16;
  int32
    affineData[6];

  if (regionMode == REGMODE_INTER_PRED_DELTA ||
      regionMode == REGMODE_INTER_NOPRED)
    mcpOrthToMvDelta(motCoefQuant, affineData, regionSize);
  else
    for(j=0; j<6; j++)
      affineData[j] = 0;

  if (regionMode == REGMODE_INTER_PRED_DELTA ||
      regionMode == REGMODE_INTER_PRED)
    for(j=0; j<6; j++)
      affineData[j] += affineDataPred[j];

  mvcBuildPredRegionYUV(region, affineData,
                        prevfr->yptr, predfr->yptr,
                        prevfr->uptr, predfr->uptr,
                        prevfr->vptr, predfr->vptr,
                        picSizeX, picSizeY);
}

double lagrSegmentCost(struct frame *prevfr, struct frame *curfr,
                       object *current_region, object *object_list,int regionIndex,
                       int *motCoefQuant, int32 *affineDataPred,
                       int picSizeX, int picSizeY,
                       double lambda, int qp, int aQP, int regionMode,
                       int predDir, int tryPec, u_char *pecModeOut,
                       int *bitsOut)
{
 
  int
    i, j,
    bitsMode, bitsMotion, bitsOvh, bitsDct, bitsDctC, regionSize,
    predSelNeeded,
    xLen, yLen;
  double
    lagrangian;

  static int
    first_invo = 1;

  Byte
    predY[256],
    predU[64],
    predV[64],
    *yPtrPrev, *yPtrPred,
    *uPtrPrev, *uPtrPred,
    *vPtrPrev, *vPtrPred;

  struct frame
    predfr;

  predfr.yptr = predY;   
  predfr.uptr = predU;   
  predfr.vptr = predV;
   
  regionSize = current_region->no_of_pels == 64 ? 8 : 16;

  if(regionMode == REGMODE_INTRA){

    return
      SendIntraDctRegion(curfr, NULL, current_region,object_list, regionIndex, picSizeX,
                         picSizeY, regionSize, current_region->predPossible,
                         pecModeOut, &bitsMode, &bitsOvh, &bitsDct, &bitsDctC,
                         qp, (float)lambda, 0, 0);

  }

  else{
    
    if(regionMode == REGMODE_INTER_NOPRED ||
       regionMode == REGMODE_INTER_PRED_DELTA){

      if(motCoefQuant[0] == 0 &&
         motCoefQuant[1] == 0 &&
         motCoefQuant[2] == 0 &&
         motCoefQuant[3] == 0 &&
         motCoefQuant[4] == 0 &&
         motCoefQuant[5] == 0){
        *pecModeOut = -1;
        return HUGE;
      }
    }

    /* Build Prediction */

    if(regionMode == REGMODE_INTER_NOPRED ||
       regionMode == REGMODE_INTER_PRED   ||
       regionMode == REGMODE_INTER_PRED_DELTA){

      lagrPredRegion(current_region, motCoefQuant, affineDataPred,
                     prevfr, &predfr, picSizeX, picSizeY, regionMode);
    }
    else{

      /* Copy region */

      /* Width of region is assumed to be multiple of 4 */
      assert(((current_region->xmax - current_region->xmin + 1) & 3) == 0 &&
             sizeof(Byte) == 1);

      yPtrPred = predfr.yptr;
      yPtrPrev = prevfr->yptr + current_region->ymin * picSizeX +
                 current_region->xmin;
      yLen = current_region->ymax - current_region->ymin;
      xLen = current_region->xmax - current_region->xmin;
      for(i = 0; i <= yLen; i++, yPtrPred += 16, yPtrPrev += picSizeX){
        for(j = 0; j <= xLen; j += 4){
          yPtrPred[j] = yPtrPrev[j];
          yPtrPred[j+1] = yPtrPrev[j+1];
          yPtrPred[j+2] = yPtrPrev[j+2];
          yPtrPred[j+3] = yPtrPrev[j+3];
        }
      }
      uPtrPred = predfr.uptr;
      uPtrPrev = prevfr->uptr + ((current_region->ymin * picSizeX) >> 2) +
                 (current_region->xmin >> 1);
      vPtrPred = predfr.vptr;
      vPtrPrev = prevfr->vptr + ((current_region->ymin * picSizeX) >> 2) +
                 (current_region->xmin >> 1);
      yLen >>= 1; xLen >>= 1;
      for(i = 0; i <= yLen; i++, uPtrPred += 8, uPtrPrev += picSizeX >> 1,
        vPtrPred += 8, vPtrPrev += picSizeX >> 1){
        for(j = 0; j <= xLen; j += 2){
          uPtrPred[j] = uPtrPrev[j];
          uPtrPred[j+1] = uPtrPrev[j+1];
          vPtrPred[j] = vPtrPrev[j];
          vPtrPred[j+1] = vPtrPrev[j+1];
        }
      }
    }

    /* Code prediction error and count bits */

    if(regionMode == REGMODE_INTER_PRED   ||
       regionMode == REGMODE_INTER_PRED_DELTA){
      if(predDir == PREDDIR_UP)
        predSelNeeded = current_region->predSelNeededUp;
      else
        predSelNeeded = current_region->predSelNeededLeft;
    }
    else
      predSelNeeded = PREDSEL_NOTNEEDED;

    lagrangian = 
      SendDctRegion(curfr, &predfr, 0, NULL, current_region, object_list,
                    picSizeX, picSizeY, regionSize,
                    regionMode, regionIndex, current_region->predPossible,
                    current_region->predDirsAvail, predDir,
                   predSelNeeded, PREDSEL_1, /* <- doesn't affect lagr cost */
                    motCoefQuant, pecModeOut,
                    &bitsMode, &bitsMotion, &bitsOvh, &bitsDct, &bitsDctC,
                    qp, aQP, (float)lambda, LAGR_USE_DCT, 0, 0);
  }

  return lagrangian;

}
